#!/usr/bin/python
from dataclasses import dataclass, asdict


@dataclass
class Namespace:
    """A namespace in which measurements get stored. A namespace determines who can
    access measurements and the retention of the data stored in the namespace.

    All Influx versions have incorporated this concept in one way or another:
    - In InfluxDB 1 the namespace is determined by both the database and the retention
      policy of a measurement, see:
      https://docs.influxdata.com/influxdb/v1.8/concepts/glossary/#retention-policy-rp
    - In InfluxDB 2 the namespace is determined by the organization and the bucket of
      a measurement, see:
      https://docs.influxdata.com/influxdb/v2.0/organizations/buckets/
    - In InfluxDB 3 the namespace is specified directly using a single string, see:
      https://github.com/influxdata/influxdb_iox#use-influxdb-20-api-compatibility

    Attaching a namespace abstracts which InfluxDB version is used for storage and thus
    allows to construct points in a consistent way over all versions.
    """

    @staticmethod
    def version() -> str:
        """Return a version string corresponding to the namespace type"""
        raise NotImplementedError(
            "This function must be implemented by all Namespace subclasses"
        )

    @staticmethod
    def create(**kwargs):
        """A method which allows to construct a concrete Namespace objects based on the
        passed keywords. If no Namespace can be constructed, a ValueError or TypeError is
        raised

        Args:
          - keyword arguments 'database' and 'retentionPolicy' for a V1Namespace
          - keyword arguments 'organization' and 'bucket' for a V2Namespace
          - keyword argument 'name' for a V3Namespace

        Returns:
          V1Namespace|V2Namespace|V3Namespace: The namespace corresponding to the
            passed keyword arguments

        Example:
        >>> from eniris.point import Namespace
        >>> Namespace.create(database='myDatabase', retentionPolicy='myRetentionPolicy')
        V1Namespace(database='myDatabase', retentionPolicy='myRetentionPolicy')
        >>> Namespace.create(organization='myOrganization', bucket='myBucket')
        V2Namespace(organization='myOrganization', bucket='myBucket')
        >>> Namespace.create(name='myNamespace')
        V3Namespace(name='myNamespace')
        """
        if "database" in kwargs and "retentionPolicy" in kwargs:
            return V1Namespace(
                database=kwargs["database"], retentionPolicy=kwargs["retentionPolicy"]
            )
        if "organization" in kwargs and "bucket" in kwargs:
            return V2Namespace(
                organization=kwargs["organization"], bucket=kwargs["bucket"]
            )
        if "name" in kwargs:
            return V3Namespace(name=kwargs["name"])
        raise ValueError("Unable to detect the namespace type")

    def toUrlParameters(self) -> "dict[str, str]":
        """A method which returns which url parameters should be attached to a
        POST request which is storing data for a namespace

        Returns:
          dict[str, str]: The url parameters required to store data in
            the specified namespace
        """
        raise NotImplementedError("This method should be overrriden in child classes")

    def toJson(self):
        """A JSON dumpable representation of the Namespace object, which can be
        converted back into the object using the Namespace.fromJson method"""
        return {**asdict(self), "version": self.version()}


@dataclass
class V1Namespace(Namespace):
    """The InfluxDB 1 implementation of the namespace concept,
    i.e. the combination of a 'database' and 'retentionPolicy'"""

    database: str
    retentionPolicy: str

    @staticmethod
    def version() -> str:
        return "1"

    @staticmethod
    def validateDatabase(database: str) -> None:
        """Check wether the argument is a valid database name

        Args:
            value: Anything really

        Returns:
            None: An exception is raised when the argument is not a valid database name
        """
        if not isinstance(database, str):
            raise TypeError("Database must be a string")
        if len(database) == 0:  # Not required by Influx, but required by Eniris
            raise ValueError("Database must have a length of at least one character")

    @property  # type: ignore
    def database(self):
        """Get the database of the namespace

        Returns:
            str
        """
        return self._database

    @database.setter
    def database(self, database: str):
        """Set the database of the namespace

        Args:
            database (str): A valid database name
        """
        V1Namespace.validateDatabase(database)
        self._database = database

    @staticmethod
    def validateRetentionPolicy(retentionPolicy: str) -> None:
        """Check wether the argument is a valid retention policy name

        Args:
            retentionPolicy: Anything really

        Returns:
            None: An exception is raised when the argument is not a valid \
              retention policy name
        """
        if not isinstance(retentionPolicy, str):
            raise TypeError("Retention policy must be a string")
        if len(retentionPolicy) == 0:  # Not required by Influx, but required by Eniris
            raise ValueError(
                "Retention policy must have a length of at least one character"
            )

    @property  # type: ignore
    def retentionPolicy(self):
        """Get the retention policy of the namespace

        Returns:
            str
        """
        return self._retentionPolicy

    @retentionPolicy.setter
    def retentionPolicy(self, retentionPolicy: str):
        """Set the retention policy of the namespace

        Args:
            retentionPolicy (str): A valid retention policy name
        """
        V1Namespace.validateRetentionPolicy(retentionPolicy)
        self._retentionPolicy = retentionPolicy

    def toUrlParameters(self):
        return {"db": self._database, "rp": self._retentionPolicy}


@dataclass
class V2Namespace(Namespace):
    """The InfluxDB 2 implementation of the namespace concept, i.e.
    the combination of an 'organization' and a 'bucket'"""

    organization: str
    bucket: str

    @staticmethod
    def version() -> str:
        return "2"

    @staticmethod
    def validateOrganization(organization: str) -> None:
        """Check wether the argument is a valid organization name

        Args:
            organization: Anything really

        Returns:
            None: An exception is raised when the argument is not a \
              valid organization name
        """
        if not isinstance(organization, str):
            raise TypeError("Organization must be a string")
        # Not required by Influx, but required by Eniris
        if len(organization) == 0:
            raise ValueError(
                "Organization must have a length of at least one character"
            )

    @property  # type: ignore
    def organization(self):
        """Get the organization of the namespace

        Returns:
            str
        """
        return self._organization

    @organization.setter
    def organization(self, organization: str):
        """Set the organization of the namespace

        Args:
            organization (str): A valid organization name
        """
        V2Namespace.validateOrganization(organization)
        self._organization = organization

    @staticmethod
    def validateBucket(bucket: str) -> None:
        """Check wether the argument is a valid bucket name

        Args:
            bucket: Anything really

        Returns:
            None: An exception is raised when the argument is not a valid bucket name
        """
        if not isinstance(bucket, str):
            raise TypeError("Bucket must be a string")
        # Not required by Influx, but required by Eniris
        if len(bucket) == 0:
            raise ValueError("Bucket must have a length of at least one character")

    @property  # type: ignore
    def bucket(self):
        """Get the bucket of the namespace

        Returns:
            str
        """
        return self._bucket

    @bucket.setter
    def bucket(self, bucket: str):
        """Set the bucket of the namespace

        Args:
            bucket (str): A valid bucket name
        """
        V2Namespace.validateBucket(bucket)
        self._bucket = bucket

    def toUrlParameters(self):
        return {"org": self._organization, "bucket": self._bucket}


@dataclass
class V3Namespace(Namespace):
    """The InfluxDB 3 implementation of the namespace concept, i.e. a simple string"""

    name: str

    @staticmethod
    def version() -> str:
        return "IOx"

    @staticmethod
    def validateName(name: str):
        """Check wether the argument is a valid namespace name

        Args:
            name: Anything really

        Returns:
            None: An exception is raised when the argument is not a valid namespace name
        """
        if not isinstance(name, str):
            raise TypeError("Name must be a string")
        # Not required by Influx, but required by Eniris
        if len(name) == 0:
            raise ValueError("Name must have a length of at least one character")

    @property  # type: ignore
    def name(self):
        """Get the name of the namespace

        Returns:
            str
        """
        return self._name

    @name.setter
    def name(self, name: str):
        """Set the name of the namespace

        Args:
            name (str): A valid namespace name
        """
        V3Namespace.validateName(name)
        self._name = name

    def toUrlParameters(self):
        return {"namespace": self._name}



from eniris.telemessage import Telemessage


class TelemessageWriter:
    """TelemessageWriters have a method writeTelemessage which receives a Telemessage.
    Typically, they will perform some operations on the message, and the write it to
    an endpoint.
    """

    def writeTelemessage(self, message: Telemessage):
        raise NotImplementedError("This method should be overridden in child classes")

    def close(self):
        raise NotImplementedError("This method should be overridden in child classes")

    def flush(self):
        """Flush any internal state, i.e. make sure that any internally stored or
        buffered messages are transmitted"""


class TelemessagePrinter(TelemessageWriter):
    """A TelemessageWriter which prints any received Telemessages. Helpful for
    development and demonstration purposes"""

    def writeTelemessage(self, message: Telemessage):
        print(f"TelemessagePrinter (data length {len(message.data)}) {message}")

    def close(self):
        pass


from eniris.point.writer.writer import (
    PointWriter,
    PointPrinter,
    PointWriterDecorator,
    PointToTelemessageWriter,
)
from eniris.point.writer.buffered import BufferedPointToTelemessageWriter, \
    NaiveBufferedPointToTelemessageWriter
from eniris.point.writer.direct import DirectPointToTelemessageWriter
from eniris.point.writer.filter import PointDuplicateFilter
from eniris.point.writer.stats import PointStatsPrinter

__all__ = [
    "PointWriter",
    "PointPrinter",
    "PointWriterDecorator",
    "PointToTelemessageWriter",
    "BufferedPointToTelemessageWriter",
    "NaiveBufferedPointToTelemessageWriter",
    "DirectPointToTelemessageWriter",
    "PointDuplicateFilter",
    "PointStatsPrinter"
]



#!/usr/bin/python
from typing import Callable, Optional
import datetime
import logging
import time
import math
from threading import RLock
from http import HTTPStatus

import requests

DEFAULT_RETRY_CODES: "set[HTTPStatus|int]" = set(
    [
        HTTPStatus.TOO_MANY_REQUESTS,
        HTTPStatus.INTERNAL_SERVER_ERROR,
        HTTPStatus.BAD_GATEWAY,
        HTTPStatus.SERVICE_UNAVAILABLE,
        HTTPStatus.GATEWAY_TIMEOUT,
    ]
)


class AuthenticationFailure(Exception):
    "Raised when failing to authentiate to the Insights API"


def printKwargs(kwargs: dict, maxChars: int = 256):
    res = ""
    for k, v in kwargs.items():
        res += f"{k}={v}, "
    if len(res) >= 2:
        res = res[-2:]
    if len(res) > maxChars:
        res = res[:maxChars] + "..."
    return res


def retryRequest(
    requestsFunction: Callable,
    path: str,
    authorizationHeaderFunction: "Callable|None" = None,
    maximumRetries: int = 4,
    initialRetryDelayS: int = 1,
    maximumRetryDelayS: int = 60,
    retryStatusCodes: "Optional[set[int|HTTPStatus]]" = None,
    retryNr: int = 0,
    **req_function_kwargs,
) -> requests.Response:
    """Execute the given requests_function with the provided req_function_kwargs
    keyword arguments. If the function fails, it will try again until the amount
    of retries has exceeded.

    Args:
        requestsFunction (Callable): Requests function to use. Can be \
          requests.session.get, requests.session.post, requests.session.put \
          or requests.session.delete
        path (str): Url to use with the requests function
        authorizationHeaderFunction (Callable, optional): A function returning a valid \
          authorization header, if None no authorization header is attached to the \
          request. Defaults to None
        maximumRetries (int, optional): How many times to try again in case of a \
          failure. Defaults to 4
        initialRetryDelayS (int, optional): The initial delay between successive \
          retries in seconds. Defaults to 1
        maximumRetryDelayS (int, optional): The maximum delay between successive \
          retries in seconds. Defaults to 60
        retryStatusCodes (set[int], optional): A set of all response code for which \
          a retry attempt must be made. Defaults to {429, 500, 503}
        retryNr (int, optional): How often the call has been tried already. \
          Defaults to 0
        req_function_kwargs (dict): Keyword arguments for the requests_function.

    Returns:
        requests.Response: HTTP response
    """
    retryStatusCodes = (
        DEFAULT_RETRY_CODES if retryStatusCodes is None else retryStatusCodes
    )
    resp: "requests.Response|Exception"
    try:
        headers: "dict[str, str]" = req_function_kwargs.get("headers", {})
        if authorizationHeaderFunction is not None:
            headers["Authorization"] = authorizationHeaderFunction()
        req_function_kwargs["headers"] = headers
        resp = requestsFunction(path, **req_function_kwargs)
    except requests.Timeout as ex:
        resp = ex
    except requests.ConnectionError as ex:
        resp = ex
    if isinstance(resp, Exception):
        if retryNr + 1 <= maximumRetries:
            respText = str(resp).replace("\n", "\\n ").replace("\r", "\\r ")
            logging.warning(
                f"Retrying request after exception: {respText}. "
                f"API call: requests.{requestsFunction.__name__}({path}, "
                f"{printKwargs(req_function_kwargs)})"
            )
            time.sleep(min(initialRetryDelayS * 2**retryNr, maximumRetryDelayS))
            resp = retryRequest(
                requestsFunction,
                path,
                authorizationHeaderFunction,
                maximumRetries,
                initialRetryDelayS,
                maximumRetryDelayS,
                retryStatusCodes,
                retryNr + 1,
                **req_function_kwargs,
            )
        else:
            raise resp
    elif resp.status_code in retryStatusCodes and retryNr + 1 <= maximumRetries:
        respText = resp.text.replace("\n", "\\n ").replace("\r", "\\r ")
        logging.warning(
            f"Retrying request after response with status code {resp.status_code} "
            f"({HTTPStatus(resp.status_code).phrase}): {respText}. "
            f"API call: requests.{requestsFunction.__name__}({path}, "
            f"{printKwargs(req_function_kwargs)})"
        )
        sleepTimeS: "float|None" = None
        if "retry-after" in resp.headers:
            try:
                retryAfterStr = resp.headers["retry-after"]
                retryAfterFloat = float(retryAfterStr)
                if not math.isfinite(retryAfterFloat):
                    raise ValueError(f"Invalid 'retry-after' header: {retryAfterStr}")
                sleepTimeS = retryAfterFloat
            except ValueError as ex:
                logging.warning(
                    f"Invalid 'retry-after' header will be ignored: {retryAfterStr}"
                )
        if sleepTimeS is None and 500 <= resp.status_code < 600:
            # In case of a server-side error without a specific retry-after header,
            # we use a 60 sleep time to give the system extra time to recover
            sleepTimeS = 60.0
        if sleepTimeS is None:
            sleepTimeS = initialRetryDelayS * 2**retryNr
        time.sleep(max(initialRetryDelayS, min(sleepTimeS, maximumRetryDelayS)))
        resp = retryRequest(
            requestsFunction,
            path,
            authorizationHeaderFunction,
            maximumRetries,
            initialRetryDelayS,
            maximumRetryDelayS,
            retryStatusCodes,
            retryNr + 1,
            **req_function_kwargs,
        )
    return resp


REFRESHTOKEN_LIFETIME_DURATION_S = 13 * 24 * 60 * 60
REFRESHTOKEN_FRESHNESS_DURATION_S = 7 * 24 * 60 * 60


class ApiDriver:
    """An easy thread-save interface to interact with the API, with get, post, put and
    delete methods in the style of the requests library
    (see: https://docs.python-requests.org/en/master/)"""

    def __init__(
        self,
        username: str,
        password: str,
        authUrl: str = "https://authentication.eniris.be",
        apiUrl: str = "https://api.eniris.be",
        timeoutS: int = 60,
        maximumRetries: int = 4,
        initialRetryDelayS: int = 1,
        maximumRetryDelayS: int = 60,
        retryStatusCodes: "Optional[set[int|HTTPStatus]]" = None,
        session: "Optional[requests.Session]" = None,
    ):
        """Constructor. You must specify at least a username and password

        Args:
            username (str, optional): Insights username
            password (str, optional): Insights password of the user
            authUrl (str, optional): Url of authentication endpoint. Defaults to \
              https://authentication.eniris.be
            apiUrl (str, optional): Url of api endpoint. Defaults to \
              https://api.eniris.be Use https://neodata-ingress.eniris.be/v1/telemetry \
              when writing telemetry
            timeoutS (int, optional): API timeout in seconds. Defaults to 60
            maximumRetries (int, optional): How many times to try again in case of a \
              failure. Defaults to 4
            initialRetryDelayS (int, optional): The initial delay between successive \
              retries in seconds. Defaults to 1
            maximumRetryDelayS (int, optional): The maximum delay between successive \
              retries in seconds. Defaults to 60
            retryStatusCodes (set[int], optional): A set of all response code for \
              which a retry attempt must be made. Defaults to {429, 500, 503}
            session (requests.Session, optional): A session object to use for all API \
              calls. If None, a requests.Session without extra options is created. \
              Defaults to None
        """
        self.username = username
        self.password = password
        self.authUrl = authUrl
        self.apiUrl = apiUrl
        self.timeoutS = timeoutS
        self.maximumRetries = maximumRetries
        self.initialRetryDelayS = initialRetryDelayS
        self.maximumRetryDelayS = maximumRetryDelayS
        self.retryStatusCodes: "set[int|HTTPStatus]" = (
            DEFAULT_RETRY_CODES if retryStatusCodes is None else retryStatusCodes
        )
        self.refreshTokenLock = RLock()
        self.refreshDtAndToken = None
        self.accessTokenLock = RLock()
        self.accessDtAndToken = None
        self.session = requests.Session() if session is None else session

    def refreshtoken(self):
        """Get a refresh token to authenticate with the API

        Returns:
            string: A refresh token of the format `Bearer token`
        """
        with self.refreshTokenLock:
            currentDt = datetime.datetime.now()
            if (
                self.refreshDtAndToken is None
                or (currentDt - self.refreshDtAndToken[0]).total_seconds()
                > REFRESHTOKEN_LIFETIME_DURATION_S
            ):  # 13 days
                data = {"username": self.username, "password": self.password}
                try:
                    resp = retryRequest(
                        self.session.post,
                        f"{self.authUrl}/auth/login",
                        json=data,
                        timeout=self.timeoutS,
                        maximumRetries=self.maximumRetries,
                        initialRetryDelayS=self.initialRetryDelayS,
                        maximumRetryDelayS=self.maximumRetryDelayS,
                        retryStatusCodes=self.retryStatusCodes,
                    )
                    if resp.status_code != 200:
                        raise AuthenticationFailure(f"Unable to login: {resp.text}")
                except requests.Timeout as ex:
                    raise AuthenticationFailure(
                        "Unable to login: the API did not respond in time"
                    ) from ex
                self.refreshDtAndToken = (currentDt, resp.text)
            elif (
                currentDt - self.refreshDtAndToken[0]
            ).total_seconds() > REFRESHTOKEN_FRESHNESS_DURATION_S:
                try:
                    # No retries here, since this is not critical...
                    resp = self.session.get(
                        f"{self.authUrl}/auth/refreshtoken",
                        headers={
                            "Authorization": f"Bearer {self.refreshDtAndToken[1]}"
                        },
                        timeout=self.timeoutS,
                    )
                    if resp.status_code == 200:
                        self.refreshDtAndToken = (currentDt, resp.text)
                    else:
                        # Not the biggest problem, sice the refresh token will still be
                        # valid for a while, but we should log an exception
                        logging.warning(
                            f"Unable to renew the refresh token: {resp.text}"
                        )
                except requests.Timeout:
                    # Not the biggest problem, sice the refresh token will still be
                    # valid for a while, but we should log an exception
                    logging.warning(
                        "Unable to renew the refresh token: "
                        "the API did not respond in time"
                    )
            return f"Bearer {self.refreshDtAndToken[1]}"

    def accesstoken(self):
        """Get an access token to authenticate with the API

        Returns:
            string: An access token of the format `Bearer token`
        """
        with self.accessTokenLock:
            currentDt = datetime.datetime.now()
            if (
                self.accessDtAndToken is None
                or (currentDt - self.accessDtAndToken[0]).total_seconds() > 3.75 * 60
            ):  # 3.75 minutes (max 5 minutes, minus some leeway)
                try:
                    resp = retryRequest(
                        self.session.get,
                        f"{self.authUrl}/auth/accesstoken",
                        authorizationHeaderFunction=self.refreshtoken,
                        timeout=self.timeoutS,
                        maximumRetries=self.maximumRetries,
                        initialRetryDelayS=self.initialRetryDelayS,
                        maximumRetryDelayS=self.maximumRetryDelayS,
                        retryStatusCodes=self.retryStatusCodes,
                    )
                    if resp.status_code != 200:
                        raise AuthenticationFailure(
                            f"Unable to collect an access token: {resp.text}"
                        )
                except requests.Timeout as ex:
                    raise AuthenticationFailure(
                        "Unable to collect an access token: "
                        "the API did not respond in time"
                    ) from ex
                self.accessDtAndToken = (currentDt, resp.text)
            return f"Bearer {self.accessDtAndToken[1]}"

    def close(self):
        """Log out from the API"""
        currentDt = datetime.datetime.now()
        if (
            self.refreshDtAndToken is None
            or (currentDt - self.refreshDtAndToken[0]).total_seconds()
            > 14 * 24 * 60 * 60
        ):  # 14 days
            # The refresh token did already expire, there is no reason to log out
            return
        try:
            resp = retryRequest(
                self.session.post,
                f"{self.authUrl}/auth/logout",
                authorizationHeaderFunction=self.refreshtoken,
                timeout=self.timeoutS,
                maximumRetries=self.maximumRetries,
                initialRetryDelayS=self.initialRetryDelayS,
                maximumRetryDelayS=self.maximumRetryDelayS,
                retryStatusCodes=self.retryStatusCodes,
            )
            if resp.status_code in (204, 401):
                # The refresh token was either succesfully added to the deny list,
                # or it was already invalid
                self.refreshDtAndToken = None
                self.accessDtAndToken = None
            else:
                raise AuthenticationFailure(f"Unable to logout: {resp.text}")
        except requests.Timeout as ex:
            raise AuthenticationFailure(
                "Unable to logout: the API did not respond in time"
            ) from ex

    def get(self, path: str, params=None, **kwargs) -> requests.Response:
        """API GET call

        Args:
            path (str): Path relative to the apiUrl or a full url
            params (dict, optional): URL parameters. Defaults to None.

        Returns:
            requests.Response: API call response
        """
        path = (
            path
            if path.startswith("http://") or path.startswith("https://")
            else f"{self.apiUrl}{path}"
        )
        return retryRequest(
            self.session.get,
            path,
            params=params,
            authorizationHeaderFunction=self.accesstoken,
            timeout=self.timeoutS,
            maximumRetries=self.maximumRetries,
            initialRetryDelayS=self.initialRetryDelayS,
            maximumRetryDelayS=self.maximumRetryDelayS,
            retryStatusCodes=self.retryStatusCodes,
            **kwargs,
        )

    def post(
        self, path: str, json=None, params=None, data=None, **kwargs
    ) -> requests.Response:
        """API POST call()

        Args:
            path (str): Path relative to the apiUrl or a full url
            json (dict, optional): JSON body. Defaults to None.
            params (dict, optional): URL parameters. Defaults to None.

        Returns:
            requests.Response: API call response
        """
        path = (
            path
            if path.startswith("http://") or path.startswith("https://")
            else f"{self.apiUrl}{path}"
        )
        return retryRequest(
            self.session.post,
            path,
            json=json,
            params=params,
            data=data,
            authorizationHeaderFunction=self.accesstoken,
            timeout=self.timeoutS,
            maximumRetries=self.maximumRetries,
            initialRetryDelayS=self.initialRetryDelayS,
            maximumRetryDelayS=self.maximumRetryDelayS,
            retryStatusCodes=self.retryStatusCodes,
            **kwargs,
        )

    def put(
        self, path: str, json=None, params=None, data=None, **kwargs
    ) -> requests.Response:
        """API PUT call

        Args:
            path (str): Path relative to the apiUrl or a full url
            json (dict, optional): JSON body. Defaults to None.
            params (dict, optional): URL parameters. Defaults to None.

        Returns:
            requests.Response: API call response
        """
        path = (
            path
            if path.startswith("http://") or path.startswith("https://")
            else f"{self.apiUrl}{path}"
        )
        return retryRequest(
            self.session.put,
            path,
            json=json,
            params=params,
            data=data,
            authorizationHeaderFunction=self.accesstoken,
            timeout=self.timeoutS,
            maximumRetries=self.maximumRetries,
            initialRetryDelayS=self.initialRetryDelayS,
            maximumRetryDelayS=self.maximumRetryDelayS,
            retryStatusCodes=self.retryStatusCodes,
            **kwargs,
        )

    def delete(self, path: str, params=None, **kwargs) -> requests.Response:
        """API DELETE call

        Args:
            path (str): Path relative to the apiUrl or a full url
            params (dict, optional): URL parameters. Defaults to None.

        Returns:
            requests.Response: API call response
        """
        path = (
            path
            if path.startswith("http://") or path.startswith("https://")
            else f"{self.apiUrl}{path}"
        )
        return retryRequest(
            self.session.delete,
            path,
            params=params,
            authorizationHeaderFunction=self.accesstoken,
            timeout=self.timeoutS,
            maximumRetries=self.maximumRetries,
            initialRetryDelayS=self.initialRetryDelayS,
            maximumRetryDelayS=self.maximumRetryDelayS,
            retryStatusCodes=self.retryStatusCodes,
            **kwargs,
        )



#!/usr/bin/python
from dataclasses import dataclass
from collections import UserDict
from datetime import datetime
import math
from typing import Mapping, Union

from eniris.point.namespace import Namespace


class TagSet(UserDict):
    """A set of measured values.
    Since a TagSet is created automatically when passing a dictionary as the 'tags'
    argument of the Point constructor, one usually does not have to instantiate
    this class directly.

    This corresponds to an InfluxDB tag set, see also:
    https://docs.influxdata.com/influxdb/v2.6/reference/key-concepts/data-elements/
    """

    def __setitem__(self, key: str, value: str):
        """Set a tag set key-value pair

        Args:
            key (str): A valid tag key
            value (str): A valid tag value
        """
        TagSet.validateKey(key)
        TagSet.validateValue(value)
        super().__setitem__(key, value)

    @staticmethod
    def validateKey(key: str):
        """Check wether the argument is a valid tag key

        Args:
            key: Anything really

        Returns:
            None: An exception is raised when the argument is not a valid tag key
        """
        if not isinstance(key, str):
            raise TypeError("Tag key must be a string")
        # Not required by Influx, but required by Eniris
        if len(key) == 0:
            raise ValueError("Tag key must have a length of at least one character")
        # The docs state: 'Lines separated by the newline character \n represent a
        # single point in InfluxDB. Line protocol is whitespace sensitive.' See:
        # https://docs.influxdata.com/influxdb/v2.7/reference/syntax/line-protocol/
        if "\n" in key:
            raise ValueError("Newline characters are not allowed in tag keys")
        # See:
        # https://docs.influxdata.com/influxdb/v2.7/reference/syntax/line-protocol/#naming-restrictions
        if key[0] == "_":
            raise ValueError("Tag key cannot start with an underscore character")

    @staticmethod
    def validateValue(value: str):
        """Check wether the argument is a valid tag value

        Args:
            value: Anything really

        Returns:
            None: An exception is raised when the argument is not a valid tag value
        """
        if not isinstance(value, str):
            raise TypeError("Tag value must be a string")
        # Not required by Influx, but required by Eniris
        if len(value) == 0:
            raise ValueError("Tag values must have a length of at least one character")
        # The docs state: 'Lines separated by the newline character \n represent a
        # single point in InfluxDB. Line protocol is whitespace sensitive.' See:
        # https://docs.influxdata.com/influxdb/v2.7/reference/syntax/line-protocol/
        if "\n" in value:
            raise ValueError("Newline characters are not allowed in tag values")

    @staticmethod
    def escapeKey(key: str):
        """Convert a tag key into its line-protocol representation, escaping
        any problematic characters
        See also:
        https://docs.influxdata.com/influxdb/v2.6/reference/syntax/line-protocol

        Args:
            key (str): A valid tag key

        Returns:
            str: The line-protocol representation of the tag key
        """
        # Not strictly required, but best to do to avoid nonsense when a tag key
        # ends with a backslash. See:
        # https://docs.influxdata.com/influxdb/v2.7/reference/syntax/line-protocol/#escaping-backslashes
        key = key.replace("\\", "\\\\")
        # See:
        # https://docs.influxdata.com/influxdb/v2.7/reference/syntax/line-protocol/#special-characters
        return key.replace(",", "\\,").replace("=", "\\=").replace(" ", "\\ ")

    @staticmethod
    def escapeValue(value: str):
        """Convert a tag value into its line-protocol representation
        See also:
        https://docs.influxdata.com/influxdb/v2.6/reference/syntax/line-protocol

        Args:
            value (str): A valid tag value

        Returns:
            str: The line-protocol representation of the tag value
        """
        # Not strictly required, but best to do to avoid nonsense when a tag value ends
        # with a backslash. See:
        # https://docs.influxdata.com/influxdb/v2.7/reference/syntax/line-protocol/#escaping-backslashes
        value = value.replace("\\", "\\\\")
        # See:
        # https://docs.influxdata.com/influxdb/v2.7/reference/syntax/line-protocol/#special-characters
        return value.replace(",", "\\,").replace("=", "\\=").replace(" ", "\\ ")

    def toLineProtocol(self):
        """Convert a tag set into its line-protocol representation
        See also:
        https://docs.influxdata.com/influxdb/v2.6/reference/syntax/line-protocol

        Returns:
            str: The line-protocol representation of the tag set
        """
        lst = [f"{TagSet.escapeKey(k)}={TagSet.escapeValue(self[k])}" for k in self]
        lst.sort()
        return ",".join(lst)


class FieldSet(UserDict):
    """A set of measured values.
    Since a FieldSet is created automatically when passing a dictionary as the 'fields'
    argument of the Point constructor, one usually does not have to instantiate this
    class directly.

    This corresponds to an InfluxDB field set, see also:
    https://docs.influxdata.com/influxdb/v2.6/reference/key-concepts/data-elements/
    """

    def __setitem__(self, key: str, value: "bool|int|float|str"):
        """Set a field set key-value pair

        Args:
            key (str): A valid tag key
            value (str): A valid tag value
        """
        FieldSet.validateKey(key)
        FieldSet.validateValue(value)
        super().__setitem__(key, value)

    @staticmethod
    def validateKey(key: str):
        """Check wether the argument is a valid field key

        Args:
            key: Anything really

        Returns:
            None: An exception is raised when the argument is not a valid field key
        """
        if not isinstance(key, str):
            raise TypeError("Field key must be a string")
        # Not required by Influx, but required by Eniris
        if len(key) == 0:
            raise ValueError("Field key must have a length of at least one character")
        # The docs state: 'Lines separated by the newline character \n represent a
        # single point in InfluxDB. Line protocol is whitespace sensitive.' See:
        # https://docs.influxdata.com/influxdb/v2.7/reference/syntax/line-protocol/
        if "\n" in key:
            raise ValueError("Newline characters are not allowed in field keys")
        # See:
        # https://docs.influxdata.com/influxdb/v2.7/reference/syntax/line-protocol/#naming-restrictions
        if key[0] == "_":
            raise ValueError("Field key cannot start with an underscore character")

    @staticmethod
    def validateValue(value: "bool|int|float|str"):
        """Check wether the argument is a valid field value

        Args:
            value: Anything really

        Returns:
            None: An exception is raised when the argument is not a valid field value
        """
        if isinstance(value, bool):
            pass
        elif isinstance(value, int):
            pass
        elif isinstance(value, float):
            if not math.isfinite(value):
                raise ValueError("Floats fust be finite")
        elif isinstance(value, str):
            # The docs state: 'Lines separated by the newline character \n represent
            # a single point in InfluxDB. Line protocol is whitespace sensitive.' See:
            # https://docs.influxdata.com/influxdb/v2.7/reference/syntax/line-protocol/
            if "\n" in value:
                raise ValueError("Newline characters are not allowed in field values")
        else:
            raise TypeError(
                f"Field value {str(value)} is of the type {str(type(value))}"
            )

    @staticmethod
    def escapeKey(key: str):
        """Convert a field key into its line-protocol representation, escaping any
        problematic characters. See also:
        https://docs.influxdata.com/influxdb/v2.6/reference/syntax/line-protocol

        Args:
            key (str): A valid field key

        Returns:
            str: The line-protocol representation of the field key
        """
        # Not strictly required, but best to do to avoid nonsense when a field key
        # ends with a backslash. See:
        # https://docs.influxdata.com/influxdb/v2.7/reference/syntax/line-protocol/#escaping-backslashes
        key = key.replace("\\", "\\\\")
        # See:
        # https://docs.influxdata.com/influxdb/v2.7/reference/syntax/line-protocol/#special-characters
        return key.replace(",", "\\,").replace("=", "\\=").replace(" ", "\\ ")

    @staticmethod
    def escapeValue(value: "bool|int|float|str"):
        """Convert a field value into its line-protocol representation
        See also:
        https://docs.influxdata.com/influxdb/v2.6/reference/syntax/line-protocol

        Args:
            value (str): A valid field value

        Returns:
            str: The line-protocol representation of the field value
        """
        if isinstance(value, bool):
            return "T" if value else "F"
        if isinstance(value, int):
            return f"{value}i"
        if isinstance(value, float):
            return str(value)
        if isinstance(value, str):
            # See:
            # https://docs.influxdata.com/influxdb/v2.7/reference/syntax/line-protocol/#special-characters
            return '"' + str(value).replace("\\", "\\\\").replace('"', '\\"') + '"'
        raise TypeError("Field value is not of a valid type")

    def toLineProtocol(self):
        """Convert a field set into its line-protocol representation
        See also:
        https://docs.influxdata.com/influxdb/v2.6/reference/syntax/line-protocol

        Returns:
            str: The line-protocol representation of the field set
        """
        lst = [f"{FieldSet.escapeKey(k)}={FieldSet.escapeValue(self[k])}" for k in self]
        return ",".join(lst)


@dataclass
class Point:
    """One or more measured values sharing a single entity and timestamp.

    A point is stored in a 'measurement' (similar to a table in Excel) which is
    stored in a namespace. Each point has a single timestamp and some tags. These
    tags identify its source as well as other categorical properties which
    might be useful when querying the data. A point must also specify at least one
    field, containing the actual measured data.

    This corresponds to an InfluxDB point, see also:
    https://docs.influxdata.com/influxdb/v2.6/reference/key-concepts/data-elements/

    Example:
      >>> from eniris.point import point
      >>>
      >>> from datetime import datetime
      >>>
      >>> ns = {'database': 'myDatabase', 'retentionPolicy': 'myRetentionPolicy'}
      >>> dt = datetime(2023, 1, 1)
      >>> p0 = Point(ns, 'homeSensors', dt, {'id': 'livingroomSensor'}, {'temp_C': 18., 'humidity_perc': 20.})
      >>> p1 = Point(ns, 'homeSensors', dt, {'id': 'saunaSensor'}, {'temp_C': 40., 'humidity_perc': 90.})
      >>> print(p0)
      Point(namespace=V1Namespace(database='myDatabase', retentionPolicy='myRetentionPolicy'), measurement='homeSensors', time=datetime.datetime(2023, 1, 1, 0, 0), tags={'id': 'livingroomSensor'}, fields={'temp_C': 18.0, 'humidity_perc': 20.0})
    """

    namespace: Namespace
    measurement: str
    time: datetime
    tags: TagSet
    fields: FieldSet

    def __init__(
        self,
        namespace: "Namespace|dict",
        measurement: str,
        time: "datetime|None",
        tags: "Union[TagSet,Mapping[str, str]]",
        fields: "FieldSet|Mapping[str, bool|int|float|str]",
    ):
        self.namespace = namespace  # type: ignore
        self.measurement = measurement
        self.time = time  # type: ignore
        self.tags = tags  # type: ignore
        self.fields = fields  # type: ignore

    @staticmethod
    def validateNamespace(namespace: Namespace):
        """Check wether the argument is a valid namespace object

        Args:
            namespace: Anything really

        Returns:
            None: An exception is raised when the argument is not a
                valid namespace object
        """
        if not isinstance(namespace, Namespace):
            raise ValueError("Namespace must be a Namespace object")

    @property  # type: ignore
    def namespace(self) -> Namespace:
        """Get the namespace of the point

        Returns:
            eniris.point.Namespace
        """
        return self._namespace

    @namespace.setter
    def namespace(self, namespace: "Namespace|dict"):
        """Set the namespace of the point

        Args:
            namespace (eniris.point.Namespace|dict): A namespace object,
                or a JSON representation of a namespace object
        """
        namespace = (
            Namespace.create(**namespace) if isinstance(namespace, dict) else namespace
        )
        Point.validateNamespace(namespace)
        self._namespace = namespace

    @staticmethod
    def validateMeasurement(measurement: str):
        """Check wether the argument is a valid measurement name

        Args:
            measurement: Anything really

        Returns:
            None: An exception is raised when the argument is not a valid
                measurement name
        """
        if not isinstance(measurement, str):
            raise ValueError("Measurement name must be a string")
        # Not required by Influx, but required by Eniris
        if len(measurement) == 0:
            raise ValueError(
                "Measurement name must have a length of at least one character"
            )
        # The docs state: 'Lines separated by the newline character \n represent a
        # single point in InfluxDB. Line protocol is whitespace sensitive.' See:
        # https://docs.influxdata.com/influxdb/v2.7/reference/syntax/line-protocol/
        if "\n" in measurement:
            raise ValueError("Newline characters are not allowed in measurement name")
        # See:
        # https://docs.influxdata.com/influxdb/v2.7/reference/syntax/line-protocol/#naming-restrictions
        if measurement[0] == "_":
            raise ValueError(
                "Measurement name cannot start with an underscore character"
            )

    @staticmethod
    def escapeMeasurement(measurement: str):
        """Convert a measurement name into its line-protocol representation, escaping
        any problematic characters. See also:
        https://docs.influxdata.com/influxdb/v2.6/reference/syntax/line-protocol

        Args:
            measurement (str): A valid measurement name

        Returns:
            str: The line-protocol representation of the measurement name
        """
        # Not strictly required, but best to do to avoid nonsense when the measurement
        # name ends with a backslash. See:
        # https://docs.influxdata.com/influxdb/v2.7/reference/syntax/line-protocol/#escaping-backslashes
        measurement = measurement.replace("\\", "\\\\")
        # See:
        # https://docs.influxdata.com/influxdb/v2.7/reference/syntax/line-protocol/#special-characters
        return measurement.replace(",", "\\,").replace(" ", "\\ ")

    @property  # type: ignore
    def measurement(self):
        """Get the measurement of the point

        Returns:
            str
        """
        return self._measurement

    @measurement.setter
    def measurement(self, measurement: str):
        """Set the measurement of the point

        Args:
            measurement (str): A string
        """
        Point.validateMeasurement(measurement)
        self._measurement = measurement

    @staticmethod
    def validateTime(time: "datetime|None"):
        """Check wether the argument is a valid datetime object

        Args:
            time: Anything really

        Returns:
            None: An exception is raised when the argument is not a
                valid datetime object
        """
        if time is not None and not isinstance(time, datetime):
            raise ValueError("Time must be either None or a datetime object")

    @property  # type: ignore
    def time(self):
        """Get the time of the point

        Returns:
            datetime
        """
        return self._time

    @time.setter
    def time(self, time: "str|datetime|None"):
        """Set the time of the point

        Args:
            time (str|datetime|None): A string in the "%Y-%m-%dT%H:%M:%S.%f%z" format,
                a datetime object or None it the timestamp should be the moment when
                the data is consumed by the receiving system
        """
        if isinstance(time, str):
            time = datetime.strptime(time, "%Y-%m-%dT%H:%M:%S.%f%z")
        Point.validateTime(time)
        self._time = time

    @property  # type: ignore
    def tags(self):
        """Get the tag set of the point

        Returns:
            eniris.point.TagSet
        """
        return self._tags

    @tags.setter
    def tags(self, tags: "TagSet|Mapping[str, str]"):
        """Set the tag set of the point

        Args:
            tags (eniris.point.TagSet|Mapping[str, str]): A TagSet object or a
                Mapping where both the keys and values are strings
        """
        self._tags = tags if isinstance(tags, TagSet) else TagSet(tags)

    @property  # type: ignore
    def fields(self):
        """Get the field set of the point

        Returns:
            eniris.point.FieldSet
        """
        return self._fields

    @fields.setter
    def fields(self, fields: "FieldSet|Mapping[str, bool|int|float|str]"):
        """Set the field set of the point

        Args:
            fields (eniris.point.FieldSet|Mapping[str, bool|int|float|str]): A TagSet
                object or a Mapping where the keys are strings and the values are
                booleans, integers, floats or strings
        """
        self._fields = fields if isinstance(fields, FieldSet) else FieldSet(fields)

    def toLineProtocol(self):
        """Convert a point into its line-protocol representation
        See also:
        https://docs.influxdata.com/influxdb/v2.6/reference/syntax/line-protocol

        Returns:
            str: The line-protocol representation of the measurement name
        """
        return (
            Point.escapeMeasurement(self._measurement)
            + ("," + self._tags.toLineProtocol() if len(self._tags) > 0 else "")
            + " "
            + self._fields.toLineProtocol()
            + (
                " " + str(int(self._time.timestamp() * 1_000_000_000))
                if self._time is not None
                else ""
            )
        )

    def toJson(self):
        """Return a JSON dumpable representation of the telemessage

        Returns:
            A dictionary with the properties 'namespace', 'measurement', 'time',
            'tags' and 'fields'.
            The time will be represented as a string in the "%Y-%m-%dT%H:%M:%S.%f%z"
            format
        """
        return {
            "namespace": self._namespace.toJson(),
            "measurement": self._measurement,
            "time": self._time.isoformat() if self._time is not None else None,
            "tags": dict(self._tags),
            "fields": dict(self._fields),
        }



import logging
import time
from datetime import datetime, timedelta, timezone
from dataclasses import dataclass
from threading import RLock, Thread, Condition
from typing import Optional, Union, Tuple, FrozenSet

from eniris.point import Point, Namespace, FieldSet
from eniris.point.writer.writer import PointToTelemessageWriter
from eniris.telemessage import Telemessage
from eniris.telemessage.writer import TelemessageWriter

# Constant to convert timestamps to nanoseconds
NANOSECOND_CONVERSION = 10**9

PointKey = Tuple[str, int, FrozenSet[Tuple[str, str]]]


def createPointKey(point: Point) -> PointKey:
    """Create a unique key for the given Point.

    Args:
        point (Point): Point for which to create the key.

    Returns:
        PointKey: A tuple containing the measurement, timestamp in nanoseconds,\
          and a frozenset of the tag key-value pairs.
    """
    return (
        point.measurement,
        int(point.time.timestamp() * NANOSECOND_CONVERSION),
        frozenset((tagKey, point.tags[tagKey]) for tagKey in point.tags),
    )


@dataclass
class PointBuffer:
    """A buffer containing points sharing a single namespace, allowing points to be
    appended and converted to Telemessages. The buffer keeps track of the total number
    of bytes of its contents when represented  in line protocol (including newline
    characters). Each PointBuffer also stores its creation datetime.

    This class is not internally thread-safe.
    """

    namespace: Namespace
    creationDt: datetime
    pointMap: "dict[PointKey, dict[str, Union[bool,float,int,str]]]"
    nrBytes: int

    def __init__(self, namespace: Namespace):
        self.namespace = namespace
        self.creationDt = datetime.now(timezone.utc)
        self.pointMap = {}
        self.nrBytes = 0

    def calculateNrExtraBytes(self, point: Point) -> int:
        """Calculate the change in the number of bytes of the buffer line protocol
        representation when a given Point would be added.

        Args:
        - point (Point): The point to evaluate.

        Returns:
        - int: The change in the number of bytes
        """
        nrExtraBytes = 0
        pointKey = createPointKey(point)
        if pointKey not in self.pointMap:
            nrExtraBytes += (
                len(Point.escapeMeasurement(point.measurement))
                + (1 + len(point.tags.toLineProtocol()) if len(point.tags) > 0 else 0)
                + (
                    1 + len(str(int(point.time.timestamp() * NANOSECOND_CONVERSION)))
                    if point.time is not None
                    else 0
                )
                + 1
            )
            existingFields = {}
        else:
            existingFields = self.pointMap[pointKey]
        newFields = point.fields
        for fieldKey in newFields:
            if fieldKey in existingFields:
                nrExtraBytes += len(FieldSet.escapeValue(newFields[fieldKey])) - len(
                    FieldSet.escapeValue(existingFields[fieldKey])
                )
            else:
                nrExtraBytes += (
                    1
                    + len(FieldSet.escapeKey(fieldKey))
                    + 1
                    + len(FieldSet.escapeValue(newFields[fieldKey]))
                )
        return nrExtraBytes

    def append(self, point: Point):
        """Append a point to the buffer. If a point with similar attributes exists,
        it will be updated.

        Args:
        - point (Point): The point to be appended.
        """
        self.nrBytes += self.calculateNrExtraBytes(point)
        pointKey = createPointKey(point)
        existingFields = self.pointMap.setdefault(pointKey, {})
        newFields = point.fields
        for fieldKey in newFields:
            existingFields[fieldKey] = newFields[fieldKey]

    def toPoints(self):
        """Convert the stored points in the buffer back to a list of Point objects.

        Returns:
        - list[Point]: A list of Point objects reconstructed from the buffer's contents.
        """
        return [
            Point(
                self.namespace,
                measurement,
                datetime.fromtimestamp(time / NANOSECOND_CONVERSION, tz=timezone.utc),
                {el[0]: el[1] for el in tagSet},
                self.pointMap[(measurement, time, tagSet)],
            )
            for (measurement, time, tagSet) in self.pointMap
        ]

    def toTelemessage(self):
        """Convert the stored points in the buffer into a Telemessage object.

        Returns:
        - Telemessage: A Telemessage representation of the points in the buffer.
        """
        return Telemessage(
            self.namespace.toUrlParameters(),
            [p.toLineProtocol().encode("utf-8") for p in self.toPoints()],
        )


class PointBufferDict:
    """A PointBufferDict manages a dictionary of PointBuffer instances,
    indexed by namespace. It ensures that each buffer stays within a defined size
    and handles flushing buffers to Telemessages.

    Args:
      maximumBatchSizeBytes (int, optional): Maximum size in bytes for internal\
        PointBuffer (i.e. each set of points sharing the same namespace).\
        Defaults to 10 MB
      maximumBufferSizeBytes (int, optional): Maximum size in bytes all buffers in the\
        entire dictionary combined. Defaults to 100 MB

    This class is thread-safe.
    """

    def __init__(
        self,
        maximumBatchSizeBytes: int = 10_000_000,
        maximumBufferSizeBytes: int = 100_000_000,
    ):
        self.maximumBatchSizeBytes = maximumBatchSizeBytes
        self.maximumBufferSizeBytes = maximumBufferSizeBytes
        self._lock = RLock()
        self._namespace2buffer: "dict[frozenset[tuple[str, str]], PointBuffer]" = {}
        self._nrBytes = 0
        self._isStopping = False
        self._newContentOrStoppingCondition: Condition = Condition(self._lock)
        self._stoppingCondition: Condition = Condition(self._lock)

    def writePoints(self, points: "list[Point]"):
        """Writes a list of points to the buffer dictionary. If any buffer becomes too
        large, it will be flushed and a Telemessage object will be created.

        Args:
        - points (list[Point]): A list of points to write to the buffer.

        Returns:
        - list[Telemessage]: A list of Telemessage objects created from flushed buffers.
        """
        messages: "list[Telemessage]" = []
        if len(points) == 0:
            return messages
        with self._lock:
            # Add all points to namespace2buffer
            for point in points:
                namespaceParameters = point.namespace.toUrlParameters()
                namespaceKey = frozenset(
                    (key, namespaceParameters[key]) for key in namespaceParameters
                )
                buffer = self._namespace2buffer.setdefault(
                    namespaceKey, PointBuffer(point.namespace)
                )
                if (
                    buffer.nrBytes > 0
                    and buffer.nrBytes + buffer.calculateNrExtraBytes(point)
                    > self.maximumBatchSizeBytes
                ):
                    messages.append(buffer.toTelemessage())
                    self._nrBytes -= buffer.nrBytes
                    buffer = PointBuffer(point.namespace)
                    self._namespace2buffer[namespaceKey] = buffer
                # To keep self._nrBytes consistent, we remove the buffer's byte size,
                # add an element to the buffer, and then add the new buffer bytesize
                self._nrBytes -= buffer.nrBytes
                buffer.append(point)
                self._nrBytes += buffer.nrBytes
            # Check whether an immediate flush is required
            if self._nrBytes > self.maximumBufferSizeBytes:
                messages += self.flush()
            else:
                self._newContentOrStoppingCondition.notify()
        return messages

    def flush(self) -> "list[Telemessage]":
        """Flushes buffers from the dictionary, creating Telemessage objects
        for each set of points sharing the same namespace.
        This method will empty the buffer.

        Returns:
        - list[Telemessage]: A list of Telemessage objects created from the
          flushed buffers.
        """
        messages: "list[Telemessage]" = []
        with self._lock:
            for buffer in self._namespace2buffer.values():
                messages.append(buffer.toTelemessage())
            self._namespace2buffer = {}
            self._nrBytes = 0
        return messages

    def stop(self):
        with self._lock:
            self._isStopping = True
            self._newContentOrStoppingCondition.notify()
            self._stoppingCondition.notify()


class BufferedPointToTelemessageWriter(PointToTelemessageWriter):
    """A PointWriter which outputs Telemessages asynchroneously.
    It combines messages sharing the same namespace in a buffer, outputting them in a
    single Telemessage, thus ensuring that Telemessages are written efficiently.
    The maximum size of individual buffers, the combined size of all buffers and the
    maximum time a point can be hold in a buffer before being written can be configured.
    This class is useful when one or more threads generate a large numer of small
    batches of Points.

    Args:
      lingerTimeS (float or None, optional): Maximum time a point can be hold in a buffer\
        before being written. When set to None, the buffer is never flushed. Defaults to 1 s
      maximumBatchSizeBytes (int, optional): Maximum size in bytes for each PointBuffer\
        batch. Defaults to 10 MB
      maximumBufferSizeBytes (int, optional): Maximum size in bytes for the entire\
        buffer. Defaults to 100 MB

    Example:
      >>> from eniris.point import Point
      >>> from eniris.point.writer import BufferedPointToTelemessageWriter
      >>> from eniris.telemessage.writer import TelemessagePrinter
      >>> from datetime import datetime
      >>> import time
      >>>
      >>> ns = {'database': 'myDatabase', 'retentionPolicy': 'myRetentionPolicy'}
      >>> pLiving0 = Point(ns, 'homeSensors', datetime(2023, 1, 1), {'id': 'livingroomSensor'}, {'temp_C': 18., 'humidity_perc': 20.})
      >>> pSauna0 = Point(ns, 'homeSensors', datetime(2023, 1, 1), {'id': 'saunaSensor'}, {'temp_C': 40., 'humidity_perc': 90.})
      >>>
      >>> writer = BufferedPointToTelemessageWriter(TelemessagePrinter(), lingerTimeS=0.1)
      >>> writer.writePoints([pLiving0])
      >>> time.sleep(0.01)
      >>> writer.writePoints([pSauna0])
      >>> time.sleep(0.1)
      TelemessagePrinter Telemessage(parameters={'db': 'myDatabase', 'rp': 'myRetentionPolicy'}, data=[b'homeSensors,id=livingroomSensor temp_C=18.0,humidity_perc=20.0 1672527600000000000', b'homeSensors,id=saunaSensor temp_C=40.0,humidity_perc=90.0 1672527600000000000'])
    """

    def __init__(
        self,
        output: TelemessageWriter,
        lingerTimeS: "float|None" = 1,
        maximumBatchSizeBytes: int = 10_000_000,
        maximumBufferSizeBytes: int = 100_000_000,
    ):
        self.closed = False
        super().__init__(output)
        self.pointBufferDict = PointBufferDict(
            maximumBatchSizeBytes, maximumBufferSizeBytes
        )
        self.daemon: "Optional[BufferedPointToTelemessageWriterDaemon]"
        if lingerTimeS is not None:
            self.daemon = BufferedPointToTelemessageWriterDaemon(
                output, self.pointBufferDict, lingerTimeS
            )
            self.daemon.start()
        else:
            self.daemon = None

    def writePoints(self, points: "list[Point]"):
        """Write each Point of a list to its the buffer corresponding with its
        namespace. If a specific buffer becomes too large or if the combined size of
        all buffers becomed too large, it will be emptied and a Telemessage will be
        sent via the output.

        Args:
        - points (list[Point]): A list of points to write to the buffer.
        """
        self._writeMessages(self.pointBufferDict.writePoints(points))

    def _flush(self):
        """Flushes all points from the namespace buffers, writing them to the output
        as Telemessages."""
        self._writeMessages(self.pointBufferDict.flush())

    def _writeMessages(self, messages: "list[Telemessage]"):
        """Write each Telemessage of a list to the output.
        When a daemon is present, messages from the deamon cannot bubble up
        and thus we catch exceptions which occur when writing from the main
        thread for consistency. When no daemon is present, exceptions are allowed
        to bubble up.

        Args:
        - messages (list[Telemessage]): A list of messages to write to the output.
        """
        for message in messages:
            if self.daemon:
                try:
                    # TODO: This is not thread safe. We shoud probably do "with self.pointBufferDict._lock:""
                    self.output.writeTelemessage(message)
                except Exception:  # pylint: disable=broad-exception-caught
                    logging.exception(
                        "Failed to write Telemessage from "
                        + "BufferedPointToTelemessageWriter"
                    )
            else:
                self.output.writeTelemessage(message)
                
    def close(self):
        """Destructor method for the BufferedPointToTelemessageWriter. Stops the
        daemon and flushes any remaining messages."""
        if not self.closed:
            self.closed = True
            self.flush()
            self.pointBufferDict.stop()
            if self.daemon:
                self.daemon.join()
            self.output.close()

    def __del__(self):
        self.close()


class BufferedPointToTelemessageWriterDaemon(Thread):
    """Daemon thread for BufferedPointToTelemessageWriter, to ensure that points are
    outputted automatically when they are stored in a buffer for a specific duration.

    This daemon ensures that buffered points are written as Telemessages even if they
    don't fill a complete buffer by checking whether the buffer holding the point is
    older than a configured duration.

    Args:
      lingerTimeS (float, optional): Maximum time a point can be hold in a buffer\
        before being written. Defaults to 1 s
    """

    # pylint: disable=protected-access

    def __init__(
        self,
        output: TelemessageWriter,
        pointBufferDict: PointBufferDict,
        lingerTimeS: float = 1,
    ):
        super().__init__()
        self.output = output
        self.lingerTimeS = lingerTimeS
        self.pointBufferDict = pointBufferDict
        self.daemon = True
        self.name = "buffered-point-to-telemsg-writer-daemon"

    def run(self) -> None:
        logging.debug("Started BufferedPointToTelemessageWriterDaemon")
        with self.pointBufferDict._lock:
            while True:
                while (
                    not self.pointBufferDict._isStopping
                    and len(self.pointBufferDict._namespace2buffer) == 0
                ):
                    self.pointBufferDict._newContentOrStoppingCondition.wait()
                if self.pointBufferDict._isStopping:
                    logging.debug("Stopped BufferedPointToTelemessageWriterDaemon")
                    return
                curDt = datetime.now(timezone.utc)
                # Empty the buffers with old content
                thresholdDt = curDt - timedelta(seconds=self.lingerTimeS)
                newNamespace2buffer = {}
                for key in self.pointBufferDict._namespace2buffer:
                    buffer = self.pointBufferDict._namespace2buffer[key]
                    if buffer.creationDt < thresholdDt:
                        try:
                            self.output.writeTelemessage(buffer.toTelemessage())
                        except Exception:  # pylint: disable=broad-exception-caught
                            logging.exception(
                                "Failed to write Telemessage from "
                                "BufferedPointToTelemessageWriterDaemon.run"
                            )
                        self.pointBufferDict._nrBytes -= buffer.nrBytes
                    else:
                        newNamespace2buffer[key] = buffer
                self.pointBufferDict._namespace2buffer = newNamespace2buffer
                # Check which buffer needs to be emptied next and sleep for an appropriate amount of time
                minCreationDt: "datetime|None" = None
                for buffer in self.pointBufferDict._namespace2buffer.values():
                    if minCreationDt is None or buffer.creationDt < minCreationDt:
                        minCreationDt = buffer.creationDt
                if minCreationDt is not None:
                    nextWakeupDt = minCreationDt + timedelta(seconds=self.lingerTimeS)
                    sleepTimeS = nextWakeupDt.timestamp() - time.time()
                    if sleepTimeS > 0:
                        self.pointBufferDict._stoppingCondition.wait(sleepTimeS)



class NaiveBufferedPointToTelemessageWriter(PointToTelemessageWriter):
    """
    A NaiveBufferedPointToTelemessageWriter is similar to a BufferedPointToTelemessageWriter,
    with the distinction that it does not guarantee that a buffer will be flushed
    immediately when the lingerTimeS has passed since its creation. Buffers will
    only be flushed if writePoints() is called, and they are expired at that moment,
    or too big in size.
    
    This is useful in cases where we want to avoid having a worker thread, and
    we are sure Points are produced regularly, and we want to ensure that points
    are written as efficiently as possible, and we don't care about the exact timing.

    Args:
      output (TelemessageWriter): The output TelemessageWriter to write to
      lingerTimeS (float, optional): Minimum time a point should be hold in a buffer\
        before being written. When set to zero, points are passed through immediately.
        Defaults to 1 s
      maximumBatchSizeBytes (int, optional): Maximum size in bytes for each PointBuffer\
        batch. Defaults to 10 MB
      maximumBufferSizeBytes (int, optional): Maximum size in bytes for the entire\
        buffer. Defaults to 100 MB

    Example:
      >>> from eniris.point import Point
      >>> from eniris.point.writer import NaiveBufferedPointToTelemessageWriter
      >>> from eniris.telemessage.writer import TelemessagePrinter
      >>> from datetime import datetime
      >>> import time
      >>>
      >>> ns = {'database': 'myDatabase', 'retentionPolicy': 'myRetentionPolicy'}
      >>> pLiving0 = Point(ns, 'homeSensors', datetime(2023, 1, 1), {'id': 'livingroomSensor'}, {'temp_C': 18., 'humidity_perc': 20.})
      >>> pSauna0 = Point(ns, 'homeSensors', datetime(2023, 1, 1), {'id': 'saunaSensor'}, {'temp_C': 40., 'humidity_perc': 90.})
      >>>
      >>> writer = NaiveBufferedPointToTelemessageWriter(TelemessagePrinter(), lingerTimeS=0.1)
      >>> writer.writePoints([pLiving0])
      >>> time.sleep(0.01)
      >>> writer.writePoints([pSauna0])
      >>> time.sleep(0.1)
      TelemessagePrinter Telemessage(parameters={'db': 'myDatabase', 'rp': 'myRetentionPolicy'}, data=[b'homeSensors,id=livingroomSensor temp_C=18.0,humidity_perc=20.0 1672527600000000000', b'homeSensors,id=saunaSensor temp_C=40.0,humidity_perc=90.0 1672527600000000000'])
    """

    def __init__(
        self,
        output: TelemessageWriter,
        lingerTimeS: float = 1.0,
        maximumBatchSizeBytes: int = 10_000_000,
        maximumBufferSizeBytes: int = 100_000_000,
    ):
        super().__init__(output)
        self.pointBufferDict = PointBufferDict(
            maximumBatchSizeBytes, maximumBufferSizeBytes
        )
        self.lingerTime = timedelta(seconds=lingerTimeS)
        self.closed = False

    def writePoints(self, points: "list[Point]"):
        """Write each Point of a list to its the buffer corresponding with its
        namespace. If a specific buffer becomes too large or if the combined size of
        all buffers becomed too large, it will be emptied and a Telemessage will be
        sent via the output.

        Args:
        - points (list[Point]): A list of points to write to the buffer.
        """
        if self.closed:
            # Drop points if the writer is closed already!
            return
        # Add points to the buffers. If they would become too large,
        # telemessages are created
        messages = self.pointBufferDict.writePoints(points)
        # Empty the buffers with old content
        thresholdDt = datetime.now(timezone.utc) - self.lingerTime
        newNamespace2buffer = {}
        with self.pointBufferDict._lock:
            for key, buffer in self.pointBufferDict._namespace2buffer.items():
                if buffer.creationDt < thresholdDt:
                    messages.append(buffer.toTelemessage())
                    self.pointBufferDict._nrBytes -= buffer.nrBytes
                else:
                    newNamespace2buffer[key] = buffer
            self.pointBufferDict._namespace2buffer = newNamespace2buffer
        # Write the messages
        self._writeMessages(messages)

    def _flush(self):
        """Flushes all points from the namespace buffers, writing them to the output
        as Telemessages."""
        self._writeMessages(self.pointBufferDict.flush())

    def _writeMessages(self, messages: "list[Telemessage]"):
        """Write each Telemessage of a list to the output.
        When a daemon is present, messages from the deamon cannot bubble up
        and thus we catch exceptions which occur when writing from the main
        thread for consistency. When no daemon is present, exceptions are allowed
        to bubble up.

        Args:
        - messages (list[Telemessage]): A list of messages to write to the output.
        """
        for message in messages:
            self.output.writeTelemessage(message)
                
    def close(self):
        """Destructor method for the BufferedPointToTelemessageWriter. Stops the
        daemon and flushes any remaining messages."""
        if not self.closed:
            self.flush()
            self.pointBufferDict.stop()
            self.closed = True
            self.output.close()

    def __del__(self):
        self.close()




from threading import RLock
import time
from collections import OrderedDict
from typing import FrozenSet, Tuple

from eniris.point import Point
from eniris.point.writer.writer import PointWriterDecorator, PointWriter

NANOSECOND_CONVERSION = 10**9
SeriesKey = Tuple[FrozenSet[Tuple[str, str]], str, FrozenSet[Tuple[str, str]], str]


class PointDuplicateFilter(PointWriterDecorator):
    """The PointDuplicateFilter class is a PointWriter providing a thread-safe,
    in-memory storage for Points that have been processed before. It is designed to
    be a filter in the data pipeline, sitting between a source of Points and another
    "PointWriter" that it sends the filtered Points to.

    For every field of an incoming Point, it creates a hashable key to represent the
    series of the field. This key includes the Point's namespace, measurement, and tags.
    It checks if this exact field value (with the same series and timestamp) was
    processed recently. Fields which were recently processed will be passed to the list
    of outgoing Point's.

    The class removes old entries from its memory to keep its size in check, based on
    the maximum entry age and the maximum count of entries per series or in total.
    If the maximum number of entries grows too large, entries are removed according to
    a LRU (Least Recently Used) strategy.

    Args:
        maximumEntryAgeS (int, optional): The maximum time a field value will be stored\
          in the memory. The age of an entry is based on the time a field was last\
          submitted to the filter, and is unrelated to time timestamp of the field\
          itself. Defaults to two days: 2*24*3600
        maximumEntryCount (int, optional): The total maximum number of entries stored\
          in the memory of the filter. Defaults to 1_000
        maximumSeriesEntryCount (int, optional): The maximum number of entries stored\
          for a single series (i.e. namespace, measurement, tagset and field name\
          combination). Defaults to 10_000_000

    Example:
      >>> from eniris.point import Point
      >>> from eniris.point.writer import PointDuplicateFilter, PointPrinter
      >>> from datetime import datetime
      >>>
      >>> ns = {'database': 'myDatabase', 'retentionPolicy': 'myRetentionPolicy'}
      >>> pLiving0 = Point(ns, 'homeSensors', datetime(2023, 1, 1), {'id': 'livingroomSensor'}, {'temp_C': 18., 'humidity_perc': 20.})
      >>> pSauna0 = Point(ns, 'homeSensors', datetime(2023, 1, 1), {'id': 'saunaSensor'}, {'temp_C': 40., 'humidity_perc': 90.})
      >>> pLiving1 = Point(ns, 'homeSensors', datetime(2023, 1, 2), {'id': 'livingroomSensor'}, {'temp_C': 18., 'humidity_perc': 20.})
      >>>
      >>> writer = PointDuplicateFilter(PointPrinter(), maximumEntryAgeS=1, maximumSeriesEntryCount=2, maximumEntryCount=4)
      >>> writer.writePoints([pLiving0])
      PointPrinter [Point(namespace=V1Namespace(database='myDatabase', retentionPolicy='myRetentionPolicy'), measurement='homeSensors', time=datetime.datetime(2023, 1, 1, 0, 0), tags={'id': 'livingroomSensor'}, fields={'temp_C': 18.0, 'humidity_perc': 20.0})]
      >>> writer.writePoints([pLiving0, pSauna0])
      PointPrinter [Point(namespace=V1Namespace(database='myDatabase', retentionPolicy='myRetentionPolicy'), measurement='homeSensors', time=datetime.datetime(2023, 1, 1, 0, 0), tags={'id': 'saunaSensor'}, fields={'temp_C': 40.0, 'humidity_perc': 90.0})]
      >>> writer.writePoints([pLiving1])
      PointPrinter [Point(namespace=V1Namespace(database='myDatabase', retentionPolicy='myRetentionPolicy'), measurement='homeSensors', time=datetime.datetime(2023, 1, 2, 0, 0), tags={'id': 'livingroomSensor'}, fields={'temp_C': 18.0, 'humidity_perc': 20.0})]
    """

    def __init__(
        self,
        output: PointWriter,
        maximumEntryAgeS=2 * 24 * 3600,
        maximumSeriesEntryCount=1_000,
        maximumEntryCount=100_000,
    ):
        super().__init__(output)
        self.maximumEntryAgeS = maximumEntryAgeS
        self.maximumEntryCount = maximumEntryCount
        self.maximumSeriesEntryCount = maximumSeriesEntryCount
        self.memory: "dict[SeriesKey, OrderedDict[int, bool|int|float|str]]" = {}
        self.entryKey2updateTs: "OrderedDict[tuple[SeriesKey, int], int]" = (
            OrderedDict()
        )
        self.lock = RLock()

    def deleteExpiredEntries(self):
        """Remove all expired entries from the filter memory."""
        with self.lock:
            thresholdTimestamp = int(
                (time.time() - self.maximumEntryAgeS) * NANOSECOND_CONVERSION
            )
            while len(self.entryKey2updateTs) > 0:
                (pointDuplicateFilterMemoryKey, pointTimestamp) = next(
                    iter(self.entryKey2updateTs)
                )
                updateTimestamp = self.entryKey2updateTs[
                    (pointDuplicateFilterMemoryKey, pointTimestamp)
                ]
                if updateTimestamp > thresholdTimestamp:
                    break
                self._delete(pointDuplicateFilterMemoryKey, pointTimestamp)

    def _delete(self, pointSeriesKey: SeriesKey, pointTimestamp: int):
        """Remove a specific entry from the filter memory.
        This method is for internal use within the class, and should only be called for
        (pointSeriesKey, pointTimestamp) which exist, as it will otherwise throw
        an Exception.

        Args:
            pointSeriesKey (SeriesKey): The namespace, measurement and tag set of the\
              entry which must be removed
            pointTimestamp (int): The timestamp of the entry which must be removed
        """
        del self.entryKey2updateTs[(pointSeriesKey, pointTimestamp)]
        pointDuplicateFilterMemoryValue = self.memory[pointSeriesKey]
        del pointDuplicateFilterMemoryValue[pointTimestamp]
        if len(pointDuplicateFilterMemoryValue) == 0:
            del self.memory[pointSeriesKey]

    def writePoints(self, points: "list[Point]"):
        """Write points to the filter output, whilst filtering out field values which
        have been pushed before by looking if the new values are identical to the stored
        values.

        Args:
            points (list[eniris.point.Point]): List of Point's

        Returns:
            None: the filtered points are passed to the output of the filter
        """
        if len(points) == 0:
            return
        out: "list[Point]" = []
        with self.lock:
            self.deleteExpiredEntries()
            currentTs = int(time.time() * NANOSECOND_CONVERSION)
            for point in points:
                if point.time is None:
                    out.append(point)
                else:
                    pTs = int(point.time.timestamp() * NANOSECOND_CONVERSION)
                    pNamespaceParameters = point.namespace.toUrlParameters()
                    pNamespaceKey = frozenset(
                        (key, pNamespaceParameters[key]) for key in pNamespaceParameters
                    )
                    pTagsKey = frozenset((key, point.tags[key]) for key in point.tags)
                    updatedFields: "dict[str, bool|int|float|str]" = {}
                    for fieldKey in point.fields:
                        seriesKey = (
                            pNamespaceKey,
                            point.measurement,
                            pTagsKey,
                            fieldKey,
                        )
                        # Add an entry for the fields of p to the data structure
                        self.entryKey2updateTs[(seriesKey, pTs)] = currentTs
                        self.entryKey2updateTs.move_to_end((seriesKey, pTs))
                        cachedSeriesValues = self.memory.setdefault(
                            seriesKey, OrderedDict()
                        )
                        # Figure out whether the field was actually updated
                        fieldValue = point.fields[fieldKey]
                        if (
                            pTs not in cachedSeriesValues
                            or cachedSeriesValues[pTs] != fieldValue
                        ):
                            cachedSeriesValues[pTs] = fieldValue
                            updatedFields[fieldKey] = fieldValue
                        # Make sure the count constraints are not violated
                        while len(cachedSeriesValues) > self.maximumSeriesEntryCount:
                            p2Ts = next(iter(cachedSeriesValues))
                            self._delete(seriesKey, p2Ts)
                        while len(self.entryKey2updateTs) > self.maximumEntryCount:
                            (p2SeriesKey, p2Ts) = next(iter(self.entryKey2updateTs))
                            self._delete(p2SeriesKey, p2Ts)
                    # If necessary, add a point to the output list
                    if len(updatedFields) > 0:
                        out.append(
                            Point(
                                point.namespace,
                                point.measurement,
                                point.time,
                                point.tags,
                                updatedFields,
                            )
                        )
        if len(out) > 0:
            self.output.writePoints(out)



from typing import Callable, Optional, Union
from http import HTTPStatus

from requests import Session

from eniris.driver import retryRequest, DEFAULT_RETRY_CODES
from eniris.telemessage import Telemessage
from eniris.telemessage.writer.writer import TelemessageWriter


class DirectTelemessageWriterUnexpectedResponse(Exception):
    "Raised when the API responded with an unexpected status code"

    def __init__(self, code, message):
        super().__init__(f"Unexpected response [code: {code}]: {message}")
        self.code = code
        self.message = message


class DirectTelemessageWriter(TelemessageWriter):
    """Write telemessages (telemetry messages) to a specific endpoint in a blocking
    fashion: using this class to write messages will block a sending thread until the
    message is succesfully transmitted or an exception is raised.
    
    Note that you will typically need to specify some optional parameters to
    succesfully authenticate.

    Args:
      url (str, optional): The url to which the Telemessages will be posted.\
        Defaults to https://neodata-ingress.eniris.be/v1/telemetry
      params (dict[str, str], optional): A dictionary with fixed parameters which\
        should be included in each request. Defaults to an empty dictionary
      authorizationHeaderFunction (Callable|None, optional): A function returning a\
        valid authorization header, if None no authorization header is attached to\
          the request. Defaults to None
      timeoutS (int, optional): Request timeout in seconds. Defaults to 60
      maximumRetries (int, optional): How many times to try again in case of a\
        failure. Defaults to 4
      initialRetryDelayS (int, optional): The initial delay between successive\
        retries in seconds. Defaults to 1
      maximumRetryDelayS (int, optional): The maximum delay between successive\
        retries in seconds. Defaults to 60
      retryStatusCodes (set[int], optional): A set of all response code for which\
        a retry attempt must be made. Defaults to {429, 500, 503}
      session (requests.Session, optional): A session object to use for all calls.\
        If None, a requests.Session without extra options is created. Defaults to None
    """

    def __init__(
        self,
        url: str = "https://neodata-ingress.eniris.be/v1/telemetry",
        params: "Optional[dict[str, str]]" = None,
        authorizationHeaderFunction: "Union[Callable,None]" = None,
        timeoutS: float = 60,
        maximumRetries: int = 4,
        initialRetryDelayS: int = 1,
        maximumRetryDelayS: int = 60,
        retryStatusCodes: "Optional[set[int|HTTPStatus]]" = None,
        session: Optional[Session] = None,
    ):
        self.url = url
        self.params = {} if params is None else params
        self.authorizationHeaderFunction = authorizationHeaderFunction
        self.timeoutS = timeoutS
        self.maximumRetries = maximumRetries
        self.initialRetryDelayS = initialRetryDelayS
        self.maximumRetryDelayS = maximumRetryDelayS
        self.retryStatusCodes: "set[int|HTTPStatus]" = (
            DEFAULT_RETRY_CODES if retryStatusCodes is None else retryStatusCodes
        )
        self.session = Session() if session is None else session
        
    def close(self):
      """There's nothing special that must be closed"""
      pass

    def writeTelemessage(self, message: Telemessage):
        """
        Write a single telemetry message to the API. This function is blocking: if it
        returns None, the message succesfully transmitted, if an unexpected response
        code is returned, a DirectTelemessageWriterUnexpectedResponse exception will be
        raised
        """
        res = retryRequest(
            self.session.post,
            self.url,
            params={**self.params, **message.parameters},
            data=message.data,
            authorizationHeaderFunction=self.authorizationHeaderFunction,
            timeout=self.timeoutS,
            maximumRetries=self.maximumRetries,
            initialRetryDelayS=self.initialRetryDelayS,
            maximumRetryDelayS=self.maximumRetryDelayS,
            retryStatusCodes=self.retryStatusCodes,
            headers=message.headers,
        )
        if res.status_code != 204:
            raise DirectTelemessageWriterUnexpectedResponse(res.status_code, res.text)



from eniris.point import Point
from eniris.point.writer.writer import PointToTelemessageWriter
from eniris.telemessage import Telemessage
from eniris.telemessage.writer import TelemessageWriter


class DirectPointToTelemessageWriter(PointToTelemessageWriter):
    """The DirectPointToTelemessageWriter class is a PointWriter which takes in Point
    objects, groups them based on their namespaces and then writes Telemessage objects
    (batched by byte size) to a provided TelemessageWriter. It makes sure that length
    of the telemessages does not exceed a configurable maximum number of bytes.

    Args:
        maximumBatchSizeBytes (int, optional): The maximum size of the combined lines\
            in a single outputted TeleMessage in bytes. Defaults to 10_000_000

    Example:
      >>> from eniris.point import Point
      >>> from eniris.point.writer import DirectPointToTelemessageWriter
      >>> from eniris.telemessage.writer import TelemessagePrinter
      >>> from datetime import datetime
      >>>
      >>> ns = {'database': 'myDatabase', 'retentionPolicy': 'myRetentionPolicy'}
      >>> pLiving0 = Point(ns, 'homeSensors', datetime(2023, 1, 1), {'id': 'livingroomSensor'}, {'temp_C': 18., 'humidity_perc': 20.})
      >>> pSauna0 = Point(ns, 'homeSensors', datetime(2023, 1, 1), {'id': 'saunaSensor'}, {'temp_C': 40., 'humidity_perc': 90.})
      >>>
      >>> writer = DirectPointToTelemessageWriter(TelemessagePrinter(), maximumBatchSizeBytes=50)
      >>> writer.writePoints([pLiving0, pSauna0])
      TelemessagePrinter Telemessage(parameters={'db': 'myDatabase', 'rp': 'myRetentionPolicy'}, data=[b'homeSensors,id=livingroomSensor temp_C=18.0,humidity_perc=20.0 1672527600000000000'])
      TelemessagePrinter Telemessage(parameters={'db': 'myDatabase', 'rp': 'myRetentionPolicy'}, data=[b'homeSensors,id=saunaSensor temp_C=40.0,humidity_perc=90.0 1672527600000000000'])
    """

    def __init__(
        self, output: TelemessageWriter, maximumBatchSizeBytes: int = 10_000_000
    ):
        super().__init__(output)
        self.maximumBatchSizeBytes = maximumBatchSizeBytes

    def writePoints(self, points: "list[Point]"):
        """Convert points to Telemessage's and writes them to the output, ensuring each
        telemessage contains points with the same namespace.

        Args:
            points (list[eniris.point.Point]): List of Point's
        """
        params2data: "dict[frozenset[tuple[str, str]], list[Point]]" = {}
        for point in points:
            namespaceParams = point.namespace.toUrlParameters()
            namespaceParamsKey = frozenset(
                (key, namespaceParams[key]) for key in namespaceParams
            )
            params2data.setdefault(namespaceParamsKey, []).append(point)

        for params, paramsData in params2data.items():
            paramsDict = {p[0]: p[1] for p in params}
            curBytes: "list[bytes]" = []
            curBytesLen = 0
            for point in paramsData:
                pBytes: bytes = point.toLineProtocol().encode("utf-8")
                if (
                    len(curBytes) != 0
                    and curBytesLen + len(pBytes) + 1 > self.maximumBatchSizeBytes
                ):  # + 1 to take into account the newlines when the lines are joined
                    self.output.writeTelemessage(Telemessage(paramsDict, curBytes))
                    curBytes = []
                    curBytesLen = 0
                curBytes.append(pBytes)
                curBytesLen += (
                    len(pBytes) + 1
                )  # Again, we do take the newline character into account
            self.output.writeTelemessage(Telemessage(paramsDict, curBytes))



from dataclasses import asdict
from datetime import datetime, timezone

from eniris.point import Point
from eniris.point.namespace import Namespace
from eniris.point.writer.writer import PointWriterDecorator, PointWriter


class PointStatsPrinter(PointWriterDecorator):
    """
    PointStatsPrinter is a PointWriterDecorator that gathers statistics about the
    points written to a PointWriter. These statistics are printed, and the points
    are send on to the output PointWriter.
    
    This class is intended to be used during debugging, to localize where data
    originates.
    """

    def __init__(
        self,
        output: PointWriter,
        origin: str,
        enablePrint:bool = False,
        printNamespace: bool = True,
        printMeasurements: bool = True,
        printTags: bool = True,
        printFields: bool = True,
        printPoints: bool = False,
        loggingNamespace: "Namespace|None" = None,
        loggingMeasurement: "str|None" = None,
    ):
        super().__init__(output)
        self.enablePrint = enablePrint
        self.printNamespace = printNamespace
        self.printMeasurements = printMeasurements
        self.printTags = printTags
        self.printFields = printFields
        self.printPoints = printPoints
        self.origin = origin
        self.loggingNamespace = loggingNamespace
        self.loggingMeasurement = loggingMeasurement

    def printStats(self, points: "list[Point]"):
        """
        Print the statistics of the points.
        """
        if self.enablePrint:
            namespace:"dict[str, int]" = {}
            measurements:"dict[str, int]" = {}
            tags:"dict[str, int]" = {}
            fields:"dict[str, int]" = {}
            for point in points:
                if isinstance(point.namespace, Namespace):
                    _namespace = asdict(point.namespace)
                elif isinstance(point.namespace, dict):
                    _namespace = point.namespace
                else:
                    raise ValueError(f"Namespace of point {point} is not of the type dict or Namespace!")
                for value in _namespace.values():
                    if value not in namespace:
                        namespace[value] = 0
                    namespace[value] += 1
                if point.measurement not in measurements:
                    measurements[point.measurement] = 0
                measurements[point.measurement] += 1
                for value in point.tags.values():
                    if value not in tags:
                        tags[value] = 0
                    tags[value] += 1
                for key in point.fields:
                    if key not in fields:
                        fields[key] = 0
                    fields[key] += 1

            msg = "-----------------------------------------------------------------\n"
            msg += "Point statistics:\n"
            msg += "\n"
            msg += f"Origin: {self.origin}\n"
            msg += f"Total number of points: {len(points)}\n"
            msg += "\n"
            if self.printNamespace:
                msg += "Per namespace value:\n"
                for key, value in namespace.items():
                    msg += f"{key}: {value} points\n"
                msg += "\n"
            if self.printMeasurements:
                msg += "Per measurement:\n"
                for key, value in measurements.items():
                    msg += f"{key}: {value} points\n"
                msg += "\n"
            if self.printTags:
                msg += "Per tag value:\n"
                for key, value in tags.items():
                    msg += f"{key}: {value} points\n"
                msg += "\n"
            if self.printFields:
                msg += f"Per field (total number of field values written: {sum(fields.values())}):\n"
                for key, value in fields.items():
                    msg += f"{key}: {value} points\n"
                msg += "\n"
            if self.printPoints:
                msg += "Points:\n"
                for point in points:
                    msg += str(point.toJson())
                msg += "\n"
            msg += "-----------------------------------------------------------------"

            # Print as a single message, as an attempt to keep everything in the console together (in case of multithreaded
            # applications the log might otherwise get scattered)
            print(msg)

    def createTrackingPoints(self, points: "list[Point]") -> "list[Point]":
        """
        Create points that track the other points, amount of data etc.
        """
        out = []
        # There is also an extensively tagged system to easily track over time in Influx
        # where the points come from, what writes them, from where in the code
        if self.loggingNamespace is not None:
            analyzed:"dict[int,tuple[dict[str,str], int]]" = {}

            for point in points:
                if isinstance(point.namespace, Namespace):
                    _namespace = asdict(point.namespace)
                elif isinstance(point.namespace, dict):
                    _namespace = point.namespace
                else:
                    raise ValueError(f"Namespace of point {point} is not of the type dict or Namespace!")

                tags = {
                    'origin': self.origin,
                    **_namespace,
                    'measurement': point.measurement,
                    **point.tags,
                }

                def signature(tags:dict):
                    hash_ = 0
                    for pair in tags.items():
                        hash_ ^= hash(pair)
                    return hash_

                sig = signature(tags)
                if sig not in analyzed:
                    analyzed[sig] = (tags, 1)
                else:
                    analyzed[sig] = (tags, analyzed[sig][1]+1)

            for item in analyzed.values():
                tags = item[0]
                count = item[1]
                print(f'count: {count}. tags: {tags}')

                out.append(
                    Point(
                        self.loggingNamespace,
                        self.loggingMeasurement,
                        datetime.now(timezone.utc),
                        tags,
                        {"count": count},
                    )
                )
                
        return out

    def writePoints(self, points: "list[Point]"):
        """Write points to the filter output, whilst filtering out field values which
        have been pushed before by looking if the new values are identical to the stored
        values.

        Args:
            points (list[eniris.point.Point]): List of Points

        Returns:
            None: the filtered points are passed to the output of the filter
        """
        self.printStats(points)
        points.extend(self.createTrackingPoints(points))
        self.output.writePoints(points)



from eniris.telemessage.writer.writer import TelemessageWriter
from eniris.telemessage import Telemessage

import gzip


class GZipTelemessageWriter(TelemessageWriter):
    """
    TelemessageWriter that compresses Telemessages using the gzip algorithm,
    and then passes them on to another TelemessageWriter. If the compression does
    not have a net positive effect in the amount of data send, then the
    original uncompressed message is sent.
    """

    def __init__(self, output: TelemessageWriter, compresslevel: int = 9):
        """Optional argument is the compression level, in range of 0-9.

        Args:
            output (TelemessageWriter): _description_
        """
        self.output = output
        self.compresslevel = compresslevel

    def writeTelemessage(self, message: Telemessage):
        gzippedData = gzip.compress(message.data, compresslevel=self.compresslevel)
        ammendedHeaders = {**message.headers, "Content-Encoding": "gzip"}
        gzippedMessage = Telemessage(
            message.parameters, gzippedData, headers=ammendedHeaders
        )

        if len(message.data) > len(gzippedData) + 23:
            # If the gzipped message is more efficient, then write this
            # 23 is the amount of bytes used for the Content-Encoding header
            self.output.writeTelemessage(gzippedMessage)
        else:
            # If it is not more efficient, then don't use compression
            self.output.writeTelemessage(message)
        
    def close(self):
        self.output.close()

    def flush(self):
        self.output.flush()



from dataclasses import dataclass, asdict
from typing import Optional


@dataclass
class Telemessage:
    """A message with lines of telemetry which must be transmitted as a single block
    (telemetry messages), including url parameters which must be included in the
    transmission"""

    parameters: "dict[str, str]"
    data: "bytes"
    headers: "dict[str,str]"

    def __init__(
        self,
        parameters: "dict[str, str]",
        data: "list[bytes]|bytes",
        headers: "Optional[dict[str, str]]" = None,
    ):
        """Constructor

        Args:
          params (dict[str, str]): Query parameters which should be added to the post\
            request in which the telemessage lines are transmitted
          data (list[bytes]|bytes): A list of bytes or bytes, where each element is a lineprotocol\
            encoded point
        """
        self.parameters = parameters
        if isinstance(data, list):
            self.data = b"\n".join(data)
        else:
            self.data = data
        if headers is None:
            self.headers = {}
        else:
            self.headers = headers

    def nrBytes(self):
        """Get the total number of bytes of the lines in the telemessage

        Returns:
            int: The total number of bytes of the lines in the telemessage
        """
        return len(self.data)

    def toJson(self):
        """Return an (almost) JSON dumpable representation of the telemessage

        Returns:
            A dictionary with a property 'parameters' and a property 'lines'
        """
        return asdict(self)



from eniris.telemessage.writer.writer import TelemessageWriter, TelemessagePrinter
from eniris.telemessage.writer.gzipped import GZipTelemessageWriter
from eniris.telemessage.writer.direct import DirectTelemessageWriter
from eniris.telemessage.writer.pooled import PooledTelemessageWriter
from eniris.telemessage.writer.background import BackgroundTelemessageWriter

__all__ = [
    "TelemessageWriter",
    "TelemessagePrinter",
    "DirectTelemessageWriter",
    "PooledTelemessageWriter",
    "GZipTelemessageWriter",
    "BackgroundTelemessageWriter"
]



from eniris.point import Point
from eniris.telemessage.writer import TelemessageWriter


class PointWriter:
    """PointWriters have a method writePoints which receives a list of Point's.
    Typically, they will perform some operations on these points and then either:
        - pass a modified list of Point's (i.e. PointWriterDecorator) to an output
        - pass a Telemessage (i.e. PointToTelemessageWriter) to an output
    """

    def writePoints(self, points: "list[Point]"):
        raise NotImplementedError("This method should be overridden in child classes")
    
    def close(self):
        pass

    def flush(self):
        pass


class PointPrinter(PointWriter):
    """A PointWriter which prints any received points. Helpful for development
    and demonstration purposes"""

    def writePoints(self, points: "list[Point]"):
        print(f"PointPrinter {points}")


class PointWriterDecorator(PointWriter):
    """A base class for PointWriters which send their output to another PointWriter.
    This class should be inherited from"""

    def __init__(self, output: PointWriter):
        self.output = output
        
    def close(self):
        self.output.close()

    def flush(self):
        """Flush any internal state, i.e. make sure that any internally stored or
        buffered points are transmitted and then flush the output"""
        self._flush()
        self.output.flush()

    def _flush(self):
        """A placeholder function which may be overridden in child classes. The
        overriding method only has to make sure that internally stored points are
        transmitted."""

    def writePoints(self, points: "list[Point]"):
        raise NotImplementedError("This method should be overridden in child classes")


class PointToTelemessageWriter(PointWriter):
    """A base class for PointWriters which send their output to a TelemessageWriter.
    This class should be inherited from"""

    def __init__(self, output: TelemessageWriter):
        self.output = output
        
    def close(self):
        self.output.close()

    def flush(self):
        """Flush any internal state, i.e. make sure that any internally stored or
        buffered messages are transmitted and then flush the output"""
        self._flush()
        self.output.flush()

    def _flush(self):
        """A placeholder function which may be overridden in child classes. The
        overriding method only has to make sure that internally stored points are
        transmitted."""

    def writePoints(self, points: "list[Point]"):
        raise NotImplementedError("This method should be overridden in child classes")



from typing import Callable, ClassVar, Optional, Union
from datetime import datetime, timezone, timedelta
from dataclasses import dataclass
from threading import Lock, RLock, Condition, Thread
from heapq import heappop, heappush
import os
import pickle
import re
import logging
import time
from http import HTTPStatus

import requests

from eniris.driver import DEFAULT_RETRY_CODES
from eniris.telemessage import Telemessage
from eniris.telemessage.writer.writer import TelemessageWriter


@dataclass
class TelemessageWrapper:
    telemessage: Telemessage
    creationDt: datetime
    subId: int
    _subIdCntLock: ClassVar[Lock] = Lock()
    _subIdCnt: ClassVar[int] = 0

    @staticmethod
    def incrementSubId():
        with TelemessageWrapper._subIdCntLock:
            TelemessageWrapper._subIdCnt += 1
            return TelemessageWrapper._subIdCnt

    def __init__(
        self,
        telemessage: Telemessage,
        creationDt: Optional[datetime] = None,
        subId: Optional[int] = None,
        scheduledDt: Optional[datetime] = None,
        retryNr: int = 0,
        snapshotPath: Optional[str] = None,
    ):
        self._lock = RLock()
        self.telemessage = telemessage
        self.creationDt = (
            datetime.now(timezone.utc) if creationDt is None else creationDt
        )
        self.subId = TelemessageWrapper.incrementSubId() if subId is None else subId
        self._scheduledDt = (
            datetime.now(timezone.utc) if scheduledDt is None else scheduledDt
        )
        self._retryNr = retryNr
        self._snapshotPath = snapshotPath
        self._finishedCondition = Condition(self._lock)
        self._isFinished = False

    @property
    def id(self):
        return (self.creationDt.isoformat(), self.subId)

    def _filename(self):
        return (
            f"{self.creationDt.strftime('%Y%m%dT%H%M%S%f')}_subId_{self.subId}"
            + f"_scheduledDt_{self._scheduledDt.strftime('%Y%m%dT%H%M%S%f')}"
            + f"_retryNr_{self._retryNr}.pickle"
        )

    def hasSnapshot(self):
        with self._lock:
            return self._snapshotPath is not None

    @staticmethod
    def loadSnapshotsFromDirectory(directory: str):
        snapshots: "list[TelemessageWrapper]" = []
        try:
            filenames = os.listdir(directory)
            for filename in filenames:
                snapshotPath = os.path.join(directory, filename)
                try:
                    telemessage = TelemessageWrapper.loadSnapshot(snapshotPath)
                    if telemessage is not None:
                        snapshots.append(telemessage)
                        logging.info(f"Loaded Telemessage from '{snapshotPath}'")
                except Exception:  # pylint: disable=broad-exception-caught
                    logging.exception(
                        f"Failed to load Telemessage from '{snapshotPath}'"
                    )
            logging.debug(f"Finished loading Telemessages from '{directory}'")
        except Exception:  # pylint: disable=broad-exception-caught
            logging.exception(f"Failed to load any Telemessages from '{directory}'")
        return snapshots

    @staticmethod
    def loadSnapshot(path: str):
        if not os.path.isfile(path):
            return None
        filename = os.path.basename(path)
        match = re.match(
            r"([^_]*)_subId_([^_]*)_scheduledDt_([^_]*)_retryNr_(\d+).pickle", filename
        )
        if not match:
            return None
        with open(path, "rb") as file:
            telemessage: Telemessage = pickle.load(file)
        creationDt = datetime.strptime(f"{match.group(1)}+00:00", "%Y%m%dT%H%M%S%f%z")
        subId = int(match.group(2))
        scheduledDt = datetime.strptime(f"{match.group(3)}+00:00", "%Y%m%dT%H%M%S%f%z")
        retryNr = int(match.group(4))
        return TelemessageWrapper(
            telemessage, creationDt, subId, scheduledDt, retryNr, path
        )

    def saveSnapshot(self, dirname: str):
        with self._lock:
            if self._snapshotPath is not None:
                return
            if self._isFinished:
                return
            try:
                snapshotPath = os.path.join(dirname, self._filename())
                with open(snapshotPath, "wb") as file:
                    pickle.dump(self.telemessage, file)
                logging.info(f"Saved Telemessage to '{snapshotPath}'")
                self._snapshotPath = snapshotPath
            except Exception:  # pylint: disable=broad-exception-caught
                logging.exception(f"Failed to save Telemessage to '{dirname}'")

    def removeSnapshot(self):
        with self._lock:
            if self._snapshotPath is None:
                return
            try:
                os.remove(self._snapshotPath)
                logging.info(f"Removed Telemessage from '{self._snapshotPath}'")
                self._snapshotPath = None
            except FileNotFoundError:
                logging.exception(
                    " ".join(
                        [
                            f"Failed to remove Telemessage from '{self._snapshotPath}'",
                            "since this file no longer exists.",
                            "Possibly another processes is modifying these files"
                        ]
                    )
                )
                self._snapshotPath = None
            except Exception:  # pylint: disable=broad-exception-caught
                logging.exception(
                    f"Failed to remove a Telemessage from '{self._snapshotPath}'"
                )

    def updateSnapshot(self):
        with self._lock:
            if self._snapshotPath is None:
                return
            if self._isFinished:
                return
            newSnapshotPath = os.path.join(
                os.path.dirname(self._snapshotPath), self._filename()
            )
            try:
                os.rename(self._snapshotPath, newSnapshotPath)
                logging.info(
                    " ".join(
                        [
                            f"Moved Telemessage from '{self._snapshotPath}'",
                            f"to '{newSnapshotPath}'",
                        ]
                    )
                )
                self._snapshotPath = newSnapshotPath
            except FileNotFoundError:
                logging.exception(
                    " ".join(
                        [
                            f"Failed to move Telemessage from '{self._snapshotPath}'",
                            f"to '{newSnapshotPath}': source file no longer exists.",
                            "Possibly another processes is modifying these files"
                        ]
                    )
                )
                self._snapshotPath = None
            except Exception:  # pylint: disable=broad-exception-caught
                logging.exception(
                    " ".join(
                        [
                            f"Failed to move Telemessage from '{self._snapshotPath}'",
                            f" to '{newSnapshotPath}'",
                        ]
                    )
                )

    def reschedule(self, reason: str, queue):
        with self._lock:
            if self._retryNr + 1 <= queue.maximumRetries:
                logging.warning(f"Retrying request after {reason}")
                self._scheduledDt = datetime.now(timezone.utc) + timedelta(
                    seconds=min(
                        queue.initialRetryDelayS * 2**self._retryNr,
                        queue.maximumRetryDelayS,
                    )
                )
                self._retryNr += 1
                self.updateSnapshot()
            else:
                logging.error(
                    " ".join(
                        [
                            "Maximum number of retries exceeded,",
                            f"dropping telemessage due to {reason}",
                        ]
                    )
                )
                self.finish(queue)
                return
        queue.addWaiting(self)

    def finish(self, queue):
        with self._lock:
            self._isFinished = True
            self._finishedCondition.notify_all()
            self.removeSnapshot()
        queue.removeActive(self)

    def isFinished(self):
        with self._lock:
            return self._isFinished

    def wait(self, timeout: "float|None" = None):
        with self._lock:
            if not self._isFinished:
                self._finishedCondition.wait(timeout)

    # Note, we assume that no other threads are trying to modify the
    # scheduledDt when comparing elements
    def __lt__(self, other):
        return self._scheduledDt < other._scheduledDt

    # Note, we assume that no other threads are trying to modify the
    # scheduledDt when comparing elements
    def __eq__(self, other):
        return self._scheduledDt == other._scheduledDt


class TelemessageWrapperQueue:
    def __init__(
        self,
        waitingMessages: "Optional[list[TelemessageWrapper]]" = None,
        maximumRetries: int = 4,
        initialRetryDelayS: int = 1,
        maximumRetryDelayS: int = 60,
    ):
        self.maximumRetries = maximumRetries
        self.initialRetryDelayS = initialRetryDelayS
        self.maximumRetryDelayS = maximumRetryDelayS
        self._lock = RLock()
        self._activeMessages: "dict[str, TelemessageWrapper]" = {}
        self._waitingMessages = [] if waitingMessages is None else list(waitingMessages)
        self._moreMessagesOrStoppingCondition: Condition = Condition(self._lock)
        self._newMessageOrStoppingCondition: Condition = Condition(self._lock)
        self._isStopping = False

    def removeActive(self, tmw: TelemessageWrapper):
        with self._lock:
            del self._activeMessages[tmw.id]

    def onWaitingToActive(self):
        with self._lock:
            while True:
                if len(self._waitingMessages) > 0:
                    tmw = self._waitingMessages[0]
                    # Since tmw._scheduledDt may only be changed while the wrapper is
                    # an active message, we may safely read the protected member
                    # _scheduledDt
                    sleepTimeS = (
                        tmw._scheduledDt.timestamp() - time.time()
                    )  # pylint: disable=protected-access
                    if sleepTimeS < 0:
                        self._moreMessagesOrStoppingCondition.notify()
                        heappop(self._waitingMessages)
                        self._activeMessages[tmw.id] = tmw
                        return tmw
                    self._moreMessagesOrStoppingCondition.wait(sleepTimeS)
                elif self._isStopping:
                    return None
                else:
                    self._moreMessagesOrStoppingCondition.wait()

    def addWaiting(self, tmw: TelemessageWrapper):
        with self._lock:
            if tmw.id in self._activeMessages:
                del self._activeMessages[tmw.id]
            else:
                self._newMessageOrStoppingCondition.notify_all()
            currentFirstWaitingMessage = (
                self._waitingMessages[0] if len(self._waitingMessages) > 0 else None
            )
            heappush(self._waitingMessages, tmw)
            # Since the protected member _scheduledDt may only be changed while
            # a wrapper is an active message, we may safely read it for both
            # tmw and currentFirstWaitingMessage
            if (
                currentFirstWaitingMessage is None
                or tmw._scheduledDt
                < currentFirstWaitingMessage._scheduledDt  # pylint: disable=protected-access
            ):
                self._moreMessagesOrStoppingCondition.notify()

    def getContentOnNewMessage(
        self, latestKnownTmw: Optional[TelemessageWrapper] = None
    ):
        with self._lock:
            while True:
                content = self.content()
                for message in content:
                    if latestKnownTmw is None or (message.creationDt, message.subId) > (
                        latestKnownTmw.creationDt,
                        latestKnownTmw.subId,
                    ):
                        return content
                if self._isStopping:
                    return None
                self._newMessageOrStoppingCondition.wait()

    def content(self):
        with self._lock:
            return self._waitingMessages + list(self._activeMessages.values())

    def stop(self):
        with self._lock:
            self._isStopping = True
            self._moreMessagesOrStoppingCondition.notify_all()
            self._newMessageOrStoppingCondition.notify_all()


class PooledTelemessageWriterDaemon(Thread):
    def __init__(
        self,
        queue: TelemessageWrapperQueue,
        url: str = "https://neodata-ingress.eniris.be/v1/telemetry",
        params: "Optional[dict[str, str]]" = None,
        authorizationHeaderFunction: "Union[Callable, None]" = None,
        timeoutS: float = 60,
        retryStatusCodes: "Optional[set[int|HTTPStatus]]" = None,
        session: Optional[requests.Session] = None,
    ):
        super().__init__()
        self.daemon = True
        self.queue = queue
        self.url = url
        self.params = params if params else {}
        self.authorizationHeaderFunction = authorizationHeaderFunction
        self.timeoutS = timeoutS
        self.retryStatusCodes: "set[int|HTTPStatus]" = (
            DEFAULT_RETRY_CODES if retryStatusCodes is None else retryStatusCodes
        )
        self.session = requests.Session() if session is None else session

    def run(self) -> None:
        logging.debug("Started PooledTelemessageWriterDaemon")
        while True:
            tmw = self.queue.onWaitingToActive()
            if tmw is None:
                logging.debug("Stopped PooledTelemessageWriterDaemon")
                return
            try:
                headers: "dict[str, str]" = (
                    {"Authorization": self.authorizationHeaderFunction()}
                    if self.authorizationHeaderFunction is not None
                    else {}
                )
                resp = self.session.post(
                    self.url,
                    params={**self.params, **tmw.telemessage.parameters},
                    data=tmw.telemessage.data,
                    headers={**headers, **tmw.telemessage.headers},
                    timeout=self.timeoutS,
                )
                if resp.status_code == 204:
                    tmw.finish(self.queue)
                elif resp.status_code in self.retryStatusCodes:
                    tmw.reschedule(
                        f"response with status code {resp.status_code} "
                        + f"({HTTPStatus(resp.status_code).phrase}): {resp.text}",
                        self.queue,
                    )
                else:
                    logging.error(
                        " ".join(
                            [
                                "Dropping telemessage due to",
                                f"response with status code {resp.status_code}",
                                f"({HTTPStatus(resp.status_code).phrase}): {resp.text}.",
                                f"Request telemessage data: {tmw.telemessage.data}",
                            ]
                        )
                    )
            except requests.Timeout:
                tmw.reschedule("timeout", self.queue)
            except requests.ConnectionError:
                tmw.reschedule("connection error", self.queue)
            except Exception:  # pylint: disable=broad-exception-caught
                logging.exception("Dropping telemessage due to unexpected exception")
                tmw.finish(self.queue)


class PooledTelemessageSnapshotDaemon(Thread):
    def __init__(
        self,
        queue: TelemessageWrapperQueue,
        snapshotFolder: str,
        minimumSnaphotAgeS: int = 5,
        maximumSnapshotStorageBytes=20_000_000,
    ):
        super().__init__()
        self.daemon = True
        self.queue = queue
        self.snapshotFolder = snapshotFolder
        self.minimumSnaphotAgeS = minimumSnaphotAgeS
        self.maximumSnapshotStorageBytes = maximumSnapshotStorageBytes

    def run(self):
        logging.debug("Started PooledTelemessageSnapshotDaemon")
        lastMessage = None
        nextMessageToBeSnapshotted = None
        while True:
            if nextMessageToBeSnapshotted is None:
                content = self.queue.getContentOnNewMessage(lastMessage)
            else:
                sleepTimeS = (
                    nextMessageToBeSnapshotted.creationDt.timestamp()
                    + self.minimumSnaphotAgeS
                    - time.time()
                )
                if sleepTimeS > 0:
                    nextMessageToBeSnapshotted.wait(sleepTimeS)
                content = self.queue.content()
            if content is None:
                logging.debug("Stopped PooledTelemessageSnapshotDaemon")
                return
            content.sort(key=lambda x: x.id)
            nextMessageToBeSnapshotted = self.fixSnaphots(content)
            if len(content) > 0:
                lastMessage = content[-1]

    def fixSnaphots(self, content: "list[TelemessageWrapper]", useAgeLimit=True):
        content.sort(key=lambda x: x.id)
        currentlySnapshottedIds = {el.id for el in content if el.hasSnapshot()}
        thresholdDt = datetime.now(timezone.utc) - timedelta(
            seconds=self.minimumSnaphotAgeS
        )
        desiredSnapshottedIds = set()
        usedBytes = 0
        for wrapper in reversed(content):
            if not useAgeLimit or wrapper.creationDt < thresholdDt:
                if (
                    usedBytes + wrapper.telemessage.nrBytes()
                    < self.maximumSnapshotStorageBytes
                ):
                    desiredSnapshottedIds.add(wrapper.id)
        for wrapper in content:
            if (
                wrapper.id in currentlySnapshottedIds
                and wrapper.id not in desiredSnapshottedIds
            ):
                wrapper.removeSnapshot()
        for wrapper in content:
            if (
                wrapper.id not in currentlySnapshottedIds
                and wrapper.id in desiredSnapshottedIds
            ):
                wrapper.saveSnapshot(self.snapshotFolder)
        for wrapper in content:
            if wrapper.creationDt >= thresholdDt:
                return wrapper
        return None


class PooledTelemessageWriter(TelemessageWriter):
    """Write telemessages (telemetry messages) to a specific endpoint
    in a non-blocking fashion: using this class to write messages will
    add the message to a LIFO (last-in-first-out) queue.
    This queue is server by a pool of worker threads, which also take
    care of message retries. To avoid dataloss in case of unexpected shutdowns,
    Messages which remain unsent can be snapshotted a specific folder.

    Args:
        url (str, optional): The url to which the Telemessages will be posted.\
            Defaults to https://neodata-ingress.eniris.be/v1/telemetry
        params (dict[str, str], optional): A dictionary with fixed parameters which\
            should be included in each request. Defaults to an empty dictionary
        authorizationHeaderFunction (Callable|None, optional): A function returning\
            a valid authorization header, if None no authorization header is\
            attached to the request. Defaults to None
        timeoutS (int, optional): Request timeout in seconds. Defaults to 60
        session (requests.Session, optional): A session object to use for all
            calls. If None, a requests.Session without extra options is created.\
            Defaults to None
        maximumRetries (int, optional): How many times to try again in case of a\
            failure. Defaults to 4
        initialRetryDelayS (int, optional): The initial delay between successive\
            retries in seconds. Defaults to 1
        maximumRetryDelayS (int, optional): The maximum delay between successive\
            retries in seconds. Defaults to 60
        retryStatusCodes (set[int], optional): A set of all response code for which\
            a retry attempt must be made. Defaults to {429, 500, 503}
    """

    def __init__(
        self,
        poolSize: int = 1,
        snapshotFolder: "str|None" = None,
        minimumSnaphotAgeS: int = 5,
        maximumSnapshotStorageBytes=20_000_000,
        url: str = "https://neodata-ingress.eniris.be/v1/telemetry",
        params: "Optional[dict[str, str]]" = None,
        authorizationHeaderFunction: "Callable|None" = None,
        timeoutS: float = 60,
        session: Optional[requests.Session] = None,
        maximumRetries: int = 4,
        initialRetryDelayS: int = 1,
        maximumRetryDelayS: int = 60,
        retryStatusCodes: "Optional[set[int|HTTPStatus]]" = None,
    ):
        self.closed = False
        params = {} if params is None else params
        retryStatusCodes = (
            DEFAULT_RETRY_CODES if retryStatusCodes is None else retryStatusCodes
        )
        waitingMessages: "list[TelemessageWrapper]" = (
            []
            if snapshotFolder is None
            else TelemessageWrapper.loadSnapshotsFromDirectory(snapshotFolder)
        )
        self.queue = TelemessageWrapperQueue(
            waitingMessages,
            maximumRetries=maximumRetries,
            initialRetryDelayS=initialRetryDelayS,
            maximumRetryDelayS=maximumRetryDelayS,
        )
        session = requests.Session() if session is None else session
        self.pool = [
            PooledTelemessageWriterDaemon(
                self.queue,
                url=url,
                params=params,
                authorizationHeaderFunction=authorizationHeaderFunction,
                timeoutS=timeoutS,
                retryStatusCodes=retryStatusCodes,
                session=session,
            )
            for i in range(poolSize)
        ]
        for deamon in self.pool:
            deamon.start()
        self.snapshotDaemon: "PooledTelemessageSnapshotDaemon|None" = None
        if snapshotFolder is not None:
            self.snapshotDaemon = PooledTelemessageSnapshotDaemon(
                self.queue,
                snapshotFolder,
                minimumSnaphotAgeS,
                maximumSnapshotStorageBytes,
            )
            self.snapshotDaemon.start()

    def writeTelemessage(self, message: Telemessage):
        """
        Write a single telemetry message to the API
        """
        self.queue.addWaiting(TelemessageWrapper(message))

    def flush(self):
        for tmw in self.queue.content():
            tmw.wait()
            
    def close(self):
        if not self.closed:
            self.queue.stop()
            for deamon in self.pool:
                deamon.join()
            if self.snapshotDaemon is not None:
                self.snapshotDaemon.join()
                # self.snapshotDaemon.fixSnaphots(self.queue.content(), False)
        self.closed = True

    def __del__(self):
        self.close()



from setuptools import setup, find_packages


with open("README.md", "r", encoding="utf8") as f:
    long_description = f.read()

VERSION = "0.8.38"
setup(
    name="eniris",
    packages=find_packages(),
    package_data={
        "eniris": ["py.typed"],
    },
    version=VERSION,
    description="Eniris API driver for Python",
    license="MIT",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Enris BV",
    author_email="info@eniris.be",
    url="https://github.com/eniris-international/eniris-api-python-driver",
    download_url=("https://github.com/eniris-international/eniris-api-python-driver/"
      f"archive/refs/tags/v{VERSION}.tar.gz"),
    keywords=["eniris", "api", "rest"],  # Keywords that define your package best
    install_requires=["requests"],
    classifiers=[
        "Development Status :: 4 - Beta",
        # Chose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable" as the
        # current state of your package
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Build Tools",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
    ],
)



from typing import Callable, ClassVar, Optional, Union
from datetime import datetime, timezone, timedelta
from dataclasses import dataclass
from threading import Lock, RLock, Thread, Event
from heapq import heappop, heappush
import logging
from http import HTTPStatus
import os
import re
import pickle
from time import time

import requests

from eniris.driver import DEFAULT_RETRY_CODES
from eniris.telemessage import Telemessage
from eniris.telemessage.writer.writer import TelemessageWriter


@dataclass
class TelemessageWrapper:
    telemessage: Telemessage
    creationDt: datetime
    subId: int
    _subIdCntLock: ClassVar[Lock] = Lock()
    _subIdCnt: ClassVar[int] = 0
    
    @staticmethod
    def incrementSubId():
        with TelemessageWrapper._subIdCntLock:
            TelemessageWrapper._subIdCnt += 1
            return TelemessageWrapper._subIdCnt

    def __init__(
        self,
        telemessage: Telemessage,
        creationDt: Optional[datetime] = None,
        subId: Optional[int] = None,
    ):
        self.telemessage = telemessage
        self.creationDt = (
            datetime.now(timezone.utc) if creationDt is None else creationDt
        )
        self._scheduledDt = self.creationDt
        self.subId = TelemessageWrapper.incrementSubId() if subId is None else subId
        self._retryNr = 0

    # Note, we assume that no other threads are trying to modify the
    # scheduledDt when comparing elements
    def __lt__(self, other):
        return self._scheduledDt < other._scheduledDt

    # Note, we assume that no other threads are trying to modify the
    # scheduledDt when comparing elements
    def __eq__(self, other):
        return self._scheduledDt == other._scheduledDt
    

class BackgroundTelemessageWriter(TelemessageWriter):
    """Write telemessages (telemetry messages) to a specific endpoint
    in a non-blocking fashion: using this class to write messages will
    add the message to a LIFO (last-in-first-out) queue.
    This queue is server by a signle worker thread, which also take
    care of message retries. To avoid dataloss in case of unexpected shutdowns,
    Messages which remain unsent can be snapshotted a specific folder.
    The snapshots will be reloaded when the service starts again.
    
    If there is a snapshotFolder and the writer is closed or destroyed, then
    all pending messages will be snapshotted at closure. If there isn't, all
    messages will be written immediately, and dropped at a failure.

    Args:
        snapshotFolder (str, optional): Folder where snapshots will be stored.\
            This must be unique for this instance! Don't let to instances write
            to the same folder.
        minimumSnaphotAgeS (float, optional): Minimum time telemessages must be\
            in memory before they are snapshotted. Defaults to 60.0 seconds.
        snapshotPeriodS (float, optional): How often the snapshot service will\
            take snapshots. If the service is shut down, this is the maximum 
            data loss period. Defaults to 3600.0 seconds.
        url (str, optional): The url to which the Telemessages will be posted.\
            Defaults to https://neodata-ingress.eniris.be/v1/telemetry
        params (dict[str, str], optional): A dictionary with fixed parameters which\
            should be included in each request. Defaults to an empty dictionary
        authorizationHeaderFunction (Callable|None, optional): A function returning\
            a valid authorization header, if None no authorization header is\
            attached to the request. Defaults to None
        timeoutS (int, optional): Request timeout in seconds. Defaults to 60
        session (requests.Session, optional): A session object to use for all
            calls. If None, a requests.Session without extra options is created.\
            Defaults to None
        maximumRetries (int, optional): How many times to try again in case of a\
            failure. Defaults to 4
        initialRetryDelayS (int, optional): The initial delay between successive\
            retries in seconds. Defaults to 1
        maximumRetryDelayS (int, optional): The maximum delay between successive\
            retries in seconds. Defaults to 60
        retryStatusCodes (set[int], optional): A set of all response code for which\
            a retry attempt must be made. Defaults to {429, 500, 503}
        maxHeapSize (int, optional): Maximum amount of items that may be on the heap.
        threadNameSuffix (str, optional): A suffix to add to the thread name.
    """

    def __init__(
        self,
        snapshotFolder: "str|None" = None,
        minimumSnaphotAgeS: float = 60.0,
        snapshotPeriodS: float = 3600.0,
        url: str = "https://neodata-ingress.eniris.be/v1/telemetry",
        params: "Optional[dict[str, str|int]]" = None,
        authorizationHeaderFunction: "Callable|None" = None,
        timeoutS: float = 60,
        session: Optional[requests.Session] = None,
        maximumRetries: int = 4,
        initialRetryDelayS: int = 1,
        maximumRetryDelayS: int = 60,
        retryStatusCodes: "Optional[set[int|HTTPStatus]]" = None,
        maxHeapSize: Optional[int] = None,
        threadNameSuffix: str = "",
        **kwargs
    ):
        self.daemon = BackgroundTelemessageWriterDaemon(
                snapshotFolder,
                minimumSnaphotAgeS,
                snapshotPeriodS,
                url,
                params,
                authorizationHeaderFunction,
                timeoutS,
                session,
                maximumRetries,
                initialRetryDelayS,
                maximumRetryDelayS,
                retryStatusCodes,
                maxHeapSize,
                threadNameSuffix,
                **kwargs
            )
        self.daemon.start()
            

    def writeTelemessage(self, message: Telemessage):
        """
        Write a single telemetry message to the API
        """
        self.daemon.writeTelemessage(message)
        
    def close(self):
        self.daemon.close()
        self.daemon.join()
        
        
    def __del__(self):
        self.close()
            
            
class BackgroundTelemessageWriterDaemon(Thread):
    
    def __init__(
        self,
        snapshotFolder: "str|None" = None,
        minimumSnaphotAgeS: float = 60.0,
        snapshotPeriodS: float = 3600.0,
        url: str = "https://neodata-ingress.eniris.be/v1/telemetry",
        params: "Optional[dict[str, str|int]]" = None,
        authorizationHeaderFunction: "Callable|None" = None,
        timeoutS: float = 60,
        session: Optional[requests.Session] = None,
        maximumRetries: int = 4,
        initialRetryDelayS: int = 1,
        maximumRetryDelayS: int = 60,
        retryStatusCodes: "Optional[set[int|HTTPStatus]]" = None,
        maxHeapSize: Optional[int] = None,
        threadNameSuffix: str = "",
        **kwargs
    ):
        super().__init__(
            name="bg-telemessage-writer" + threadNameSuffix
        )
        
        self.url = url
        self.authorizationHeaderFunction = authorizationHeaderFunction
        self.session = requests.Session() if session is None else session
        self.timeoutS = timeoutS
        self.params = {} if params is None else params
        self.retryStatusCodes: "set[int|HTTPStatus]" = (
            DEFAULT_RETRY_CODES if retryStatusCodes is None else retryStatusCodes
        )
        
        self.maximumRetries = maximumRetries
        self.initialRetryDelayS = initialRetryDelayS
        self.maximumRetryDelayS = maximumRetryDelayS
        
        self._lock = RLock()
        self._stop_signal = Event()
        self._has_new_messages_or_stop = Event()        # At a stop, both self._stop and self._has_new_messages_or_stop are set!
        self._max_heap_size = maxHeapSize
        
        self._snapshot_folder = snapshotFolder                
        self._min_snapshot_age_s = minimumSnaphotAgeS
        self._snapshot_period_s = snapshotPeriodS
        self._next_snapshot_moment = time() + self._snapshot_period_s
        if snapshotFolder is None:
            self._new_messages: "list[TelemessageWrapper]" = []
        else:
            self._new_messages: "list[TelemessageWrapper]" = self.__loadSnapshots(snapshotFolder)
            if len(self._new_messages) > 0:
                self._has_new_messages_or_stop.set()
            
        self._pending_messages:"list[TelemessageWrapper]" = []      # This will function as our heap
        self._no_messages_left = Event()
        self._no_messages_left.set()
        
        
            
    def writeTelemessage(self, message: Telemessage):
        """
        Write a single telemetry message to the API
        """
        with self._lock:
            self._new_messages.append(TelemessageWrapper(message))
            self._has_new_messages_or_stop.set()
            self._no_messages_left.clear()


    def flush(self):
        """Wait until there are no pending messages left"""
        self._no_messages_left.wait()
        
        
    def close(self, blocking:bool=False):
        """
        Closes the BackgroundTelemessageWriter.
        Pending messages will be flushed.
        After closing the BackgroundTelemessageWriter will no longer send messages.
        """
        self._stop_signal.set()        # Set the stop event first!
        self._has_new_messages_or_stop.set()
        if blocking:
            self._no_messages_left.wait()
        
    
    def stop(self):
        """A method to stop the daemon thread."""
        self.close()
        
        
    def __del__(self):
        self.close()
   
        
    def __loadSnapshots(self, snapshot_folder:str) -> "list[TelemessageWrapper]":
        if not os.path.exists(snapshot_folder):
            return []
        try:
            existing_snapshot_filenames = set(os.listdir(snapshot_folder))
        except Exception as e:
            logging.error(f"Failed reading content of snapshot folder {snapshot_folder}!")
            return []
        
        result:"list[TelemessageWrapper]" = []
        for filename in existing_snapshot_filenames:
            match = re.match(
                r"([^_]*)_subId_([^_]*).pickle", filename
            )
            if not match:
                continue
            snapshotPath = os.path.join(snapshot_folder, filename)
            try:
                with open(snapshotPath, "rb") as file:
                    telemessage: Telemessage = pickle.load(file)
                creationDt = datetime.strptime(f"{match.group(1)}+00:00", "%Y%m%dT%H%M%S%f%z")
                subId = int(match.group(2))
                result.append(TelemessageWrapper(telemessage, creationDt, subId))
            except Exception as e:
                logging.error(f"Failed loading snapshot {snapshotPath}. Exception: {e}")
        return result
    
         
    def run(self):
        while True:
            # Get the next message
            tmw = self.__get_next()
            # tmw will be None if everything has been send and there is a "stop" signal.
            # So write the pending messages to disk and end the worker
            if tmw is None:
                self.__take_snapshot()
                return
            # Try sending it to the database
            failure_reason, failed_tmw = self.__send(tmw)
            # Reschedule failed sends to a later moment
            if failed_tmw is not None:
                self.__reschedule(str(failure_reason), failed_tmw)
            # Make sure the heap doesn't become too big
            self.__lazy_limit_heap_size()
            # Create snapshots if required
            if time() > self._next_snapshot_moment:
                self.__take_snapshot()
                self._next_snapshot_moment = time() + self._snapshot_period_s
            # Signal if there are no more pending messages
            with self._lock:
                if (len(self._pending_messages) + len(self._new_messages)) == 0:
                    self._no_messages_left.set()
            
            
    def __send(self, tmw:TelemessageWrapper) -> "tuple[str, TelemessageWrapper]|tuple[None, None]":
        """
        Send the Telemessage to the database.
        If the sending fails, then the reason and the telemessage wrapper is returned.
        Otherwise None, None is returned.
        """
        try:
            headers: "dict[str, str]" = (
                {"Authorization": self.authorizationHeaderFunction()}
                if self.authorizationHeaderFunction is not None
                else {}
            )
            resp = self.session.post(
                self.url,
                params={**self.params, **tmw.telemessage.parameters},
                data=tmw.telemessage.data,
                headers={**headers, **tmw.telemessage.headers},
                timeout=self.timeoutS,
            )
            if resp.status_code == 204:
                return None, None
            elif resp.status_code in self.retryStatusCodes:
                return (
                    f"response with status code {resp.status_code} "
                    + f"({HTTPStatus(resp.status_code).phrase}): {resp.text}",
                    tmw,
                )
            else:
                logging.error(
                    " ".join(
                        [
                            "Dropping telemessage due to",
                            f"response with status code {resp.status_code}",
                            f"({HTTPStatus(resp.status_code).phrase}): {resp.text}.",
                            f"Request telemessage data: {tmw.telemessage.data}",
                        ]
                    )
                )
                return None, None
        except requests.Timeout:
            return ("timeout", tmw)
        except requests.ConnectionError:
            return ("connection error", tmw)
        except Exception:  # pylint: disable=broad-exception-caught
            logging.exception("Dropping telemessage due to unexpected exception")
            return None, None
            
            
    def __get_next(self) -> "TelemessageWrapper|None":
        """
        Get the next TelemessageWrapper to send to the database.
        This function blocks until a) a new message arrives or b) the schedule time of the first pending message happens.
        We assume that new messages are scheduled to be sent immediately (should be the case!), so they'll always bubble
        to the top of the heap immediately.
        
        This function returns the next TelemessageWrapper to be sent, except when a stop signal is received. In that case 
        it returns None.
        """
        if len(self._pending_messages) == 0:
            # If there are no more messages pending, then wait until new messages arrive
            wait_timeout_s = None
        else:
            # Else wait at most until the next pending message is scheduled to be send
            wait_timeout_s = max(0.0, (self._pending_messages[0]._scheduledDt - datetime.now(timezone.utc)).total_seconds())
        while True:
            
            if not self._stop_signal.is_set():
                # Wait until new messages arive or until the timeout has exceeded
                # - but only if there is no stop event
                self._has_new_messages_or_stop.wait(wait_timeout_s)
                
            # Put all new messages (if any) on the heap
            with self._lock:
                for tmw in self._new_messages:
                    heappush(self._pending_messages, tmw)
                self._new_messages = []
                self._has_new_messages_or_stop.clear()
                
            # If a stop event is set:
            if self._stop_signal.is_set():
                if self._snapshot_folder is None:
                    # In case of a stop event when there is no snapshot folder, try writing each pending message exactly once.
                    if len(self._pending_messages) == 0:
                        return None
                    tmw = heappop(self._pending_messages)
                    tmw._retryNr = self.maximumRetries      # This forces that there will only be one try, no reschedule.
                    return tmw
                else:
                    # Else returning None signals to snapshot all pending messages
                    return None
            elif len(self._pending_messages) > 0:
                # if not self._stop.is_set(), then there should always be pending messages - but better safe than sorry so check!
                # Get the message that is scheduled to be send first
                return heappop(self._pending_messages)
            else:
                # There are no pending messages, so wait until new messages arrive
                wait_timeout_s = None
    
    
    def __reschedule(self, reason:str, tmw:TelemessageWrapper):
        """ Reschedule sending the telemessage to a later moment - if possible - otherwise it is dropped. """
        if tmw._retryNr + 1 <= self.maximumRetries:
            logging.warning(f"Retrying request after {reason}")
            self._scheduledDt = datetime.now(timezone.utc) + timedelta(
                seconds=min(
                    self.initialRetryDelayS * 2**tmw._retryNr,
                    self.maximumRetryDelayS,
                )
            )
            tmw._retryNr += 1
            heappush(self._pending_messages, tmw)
        else:
            logging.error(
                " ".join(
                    [
                        "Maximum number of retries exceeded,",
                        f"dropping telemessage due to {reason}",
                    ]
                )
            )
            
            
    def __lazy_limit_heap_size(self):
        """ Limit the amount of messages that still has to be send. """
        # We know from the docs of heapq that:
        # "Heaps are arrays for which a[k] <= a[2*k+1] and a[k] <= a[2*k+2] for all k, counting elements from 0"
        # There is no guarantee that this is perfectly sorted from early to late in our case, but at least messages
        # at the end of the queue will tend to be later. Our lazy approach to limit the size of the heap is to drop
        # messages that are at the end of the heap. This is fast and easy.
        if self._max_heap_size is None:
            return
        while len(self._pending_messages) > self._max_heap_size:
            self._pending_messages.pop()
            
            
    def __take_snapshot(self):
        """
        Take a snapshot of all the telemessages that are currently loaded in memory, so that they are preserved
        in case the application restarts. Only telemessages that are sufficiently old are snapshotted.
        If a telemessage doesn't exist any more in memory (either it has been written to the database or it has been discarted),
        then its snapshot is also removed.
        """
        if self._snapshot_folder is None:
            return
        
        # Make sure the snapshot folder exists
        try:
            os.makedirs(self._snapshot_folder, exist_ok=True)
        except Exception as e:
            logging.error(f"Failed creating snapshot folder {self._snapshot_folder}! Not taking snapshots")
            return
        
        # Check which snapshots exist
        existing_snapshot_filenames = set(os.listdir(self._snapshot_folder))
        used_snapshot_filenames = set()
        
        # Write all telemessages to the snapshot folder if they haven't been written yet.
        dt_limit = datetime.now(timezone.utc) - timedelta(seconds=self._min_snapshot_age_s)
        
        for tmw in self._pending_messages:
            if tmw.creationDt > dt_limit:
                continue
            filename = f"{tmw.creationDt.strftime('%Y%m%dT%H%M%S%f')}_subId_{tmw.subId}.pickle"
            used_snapshot_filenames.add(filename)
            # Don't snapshot messages that are already snapshotted
            if filename in existing_snapshot_filenames:
                continue
            snapshotPath = os.path.join(self._snapshot_folder, filename)
            try:
                with open(snapshotPath, "wb") as file:
                    pickle.dump(tmw.telemessage, file)
                logging.info(f"Saved Telemessage to '{snapshotPath}'")
            except Exception as e:
                logging.error(f"Error while saving Telemessage to '{snapshotPath}'!"
                              f"Exception: {e}")
            except:
                logging.error(f"Error while saving Telemessage to '{snapshotPath}'!"
                              "Check file permissions and the existence of the snapshot folder.")
            
        # Remove all telemessages from the snapshot folder that are no longer in memory.
        obsolete_snapshot_filenames = existing_snapshot_filenames.difference(used_snapshot_filenames)
        for filename in obsolete_snapshot_filenames:
            snapshotPath = os.path.join(self._snapshot_folder, filename)
            try:
                os.remove(snapshotPath)
            except Exception as e:
                logging.error(f"Error while removing snapshot {snapshotPath}. Exception: {e}")
            except:
                logging.error(f"Unknown error while removing snapshot {snapshotPath}.")
                